#!/usr/bin/python
## -*- coding: utf-8 -*-
## 
## (C) 2008, 2013 Muthiah Annamalai
## Licensed under GPL Version 3
##
## This module is the theme file for the tamil language.
## It contains classes @Attrib, @XsyTheme
## 

colors = { u"Blue" : u"0000FF",
           u"GreenYellow" : u"ADFF2F",
           u"CornSilk":"FF8DC",
           u"IndianRed":"CD5C5C",
           u"DarkRed" : u"8B0000",
           u"Red":"FF0000",
           u"Green": u"00FF00",
           u"Coffee":"C0FFEE",
           u"Deadbe":"DEADBE",
}

class Attrib(list):
    def __init__(self,*args):
        assert( len(args) >= 2 )
        self.type = args[0]
        if ( len(args[1:]) > 1 ):
            self.extend(args[1:])
        else:
            self.append(args[1])

    @staticmethod
    def process(*args):
        u""" default process string to return inputs u"""
        return args[0]

class StringAttrib(Attrib):
    @staticmethod
    def process(*args):
        """ chicken wrapped bacon """
        return u"\""+args[0]+u"\""

class XsyTheme:
    def __init__(self):
        self.Keywords = Attrib(u'Keywords',colors["Blue"])
        self.LiteralString = StringAttrib(u'LiteralString',colors["IndianRed"])
        self.LiteralNumber = Attrib(u'LiteralNumber',colors["CornSilk"])
        self.Builtins = Attrib(u'Builtins',colors["DarkRed"])
        self.Variables = Attrib(u'Variables',colors["Green"])
        self.Operators = Attrib(u'Operators',colors["Red"])
        self.Comment = StringAttrib(u'Comment',colors["Coffee"])



#!/usr/bin/python
##
## (C) 2007, 2008-2013, 2015 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## tamil parser & AST builder - tamil frontend and  AST elements to build the parse tree.
## refactored left-recursion in the examples

import sys

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

## scanner for tamil language
from .tamil_scanner import tamilToken, tamilLex, tamilLexeme

## exceptions
from .errors import RuntimeException, ParseException

## runtime elements
from .runtime import  Environment, BuiltinFunction, \
 BlindBuiltins, DebugUtils

## AST elements
from .ast import Expr, UnaryExpr, ExprCall, ExprList, Stmt, ReturnStmt, \
 BreakStmt, ContinueStmt, ElseStmt, IfStmt, WhileStmt, DoWhileStmt, \
 ForStmt, AssignStmt, PrintStmt, DeclarationStmt, EvalStmt, ArgList, \
 ImportStmt, ValueList, Function, StmtList, Identifier, Number, \
 String, Array, Dict

## use exprs language parser
from .ExprsParser import Parser

## Tamil messages
from .tamil_messages import get_message, Messages

## Parser implementes the grammar for 'exprs' language.
## Entry point is parse(), after appropriate ctor-setup.
class tamilParser(Parser):
    """ when you add new language feature, add a AST class 
    and its evaluate methods. Also add a parser method """
    def __init__(self,lexer,fcn_map, builtin_map, dbg = False):
        if ( not isinstance(lexer, tamilLex) ):
                exception_msg = get_message(Messages.ClassNotFound)
                raise RuntimeException(Messages.ClassNotFound,u"tamil lexer")
        Parser.__init__(self,lexer,fcn_map,builtin_map,dbg)
        self.open_if_stmts = 0
        self.backtrack_atexpr = None

    @staticmethod
    def factory(lexer,fcn_map,builtin_map, dbg = False):
        """ Factory method """
        return tamilParser(lexer,fcn_map,builtin_map, dbg)

    def match(self,kind):
        ## if match return token, else ParseException
        tok = self.dequeue()
        if ( tok.kind != kind ):
            raise ParseException(u"cannot find token "+ \
                                 tamilToken.get_name(kind) + u" got " \
                                + str(tok)  \
                                + u" instead!")
        return tok

    def parse(self):
        """ parser routine """
        self.ast = StmtList(istoplevel=True)
        self.dbg_msg(u" entering parser " )
        while ( not self.lex.end_of_tokens() ):
            self.dbg_msg( u"AST length = %d"%len(self.ast) )
            if ( self.lex.peek().kind ==  tamilToken.DEF ):
                self.dbg_msg ( u"parsing for function" )
                ## save function in a global table.
                func = self.function()
                self.warn_function_overrides(func.name)
                self.function_map[func.name]=func
                self.ast.append(DeclarationStmt(func)) #add to AST
            else:
                self.dbg_msg( u"parsing for stmt" )
                st = self.stmtlist()
                if ( not self.parsing_function ):
                    self.ast.append(st)
        return self.ast

    def stmtlist(self,pass_in_ATexpr=None):
        """ parse a bunch of statements """
        self.dbg_msg(u" STMTLIST ")
        stlist = StmtList()
        while( not self.lex.end_of_tokens() ):
            self.dbg_msg(u"STMTLIST => STMT")
            ptok = self.peek()
            self.dbg_msg(u"STMTLIST "+str(ptok))
            if ( self.debug ): print(u"peek @ ",str(ptok))
            if ( ptok.kind == tamilToken.END ):
                self.dbg_msg(u"End token found");
                break
            elif ( ptok.kind == tamilToken.DOWHILE ):
                if ( self.debug ): print("DOWHILE token found")
                break            
            elif( self.inside_if and 
                 ( ptok.kind ==  tamilToken.ELSE
                   or ptok.kind == tamilToken.ATRATEOF 
                   or ptok.kind == tamilToken.CASE 
                   or ptok.kind == tamilToken.OTHERWISE ) ):
                break
            elif( ptok.kind ==  tamilToken.DEF ):
                break
            st = self.stmt(pass_in_ATexpr)
            pass_in_ATexpr = None
            stlist.append( st )
        return stlist
    
    def parseSwitchStmt(self,exp):
        ## @ <ID/EXPR> SWITCH @( expr ) CASE {stmtlist} @( expr ) CASE {stmtlist} OTHERWISE {stmtlist} END
        ## implement as an if-elseif-else statement
        self.dbg_msg("parsing SWITCH statement")
        sw_tok = self.dequeue()
        [l,c]=sw_tok.get_line_col()
        self.inside_if = True
        lhs=exp[0]
        # enter this if-statement always
        ifstmt = IfStmt( Number(1), None, None, l, c, self.debug)
        self.if_stack.append(ifstmt)
        self.dbg_msg("parsing SWITCH-body") #self.dbg_msg        
        ptok = self.peek()
        equality_token = tamilLexeme("=",tamilToken.EQUALITY)
        while ( ptok.kind == tamilToken.ATRATEOF or ptok.kind == tamilToken.OTHERWISE ):
            self.inside_if = True
            [l,c]=ptok.get_line_col()
            if ( ptok.kind == tamilToken.ATRATEOF ):
                # parse elseif branch
                self.dbg_msg("parsing CASE")
                self.match( tamilToken.ATRATEOF )
                exp = self.valuelist();
                self.dbg_msg("parsing CASE EXPR")
                self.match( tamilToken.CASE )
                next_stmt = self.stmtlist()                
                expr = Expr( lhs, equality_token, exp[0], l, c, self.debug )
                self.dbg_msg("building an Expr "+str(expr))                
                if not ifstmt.body :
                    ifstmt.expr = expr
                    ifstmt.body = next_stmt
                else:
                    case_stmt = IfStmt(expr,next_stmt,None,l,c,self.debug);
                    ifstmt.append_stmt( case_stmt )
            elif ( ptok.kind == tamilToken.OTHERWISE ):
                #parse else branch
                self.dbg_msg("parsing OTHERWISE: ")
                self.match( tamilToken.OTHERWISE )
                self.dbg_msg("parsing OTHERWISE-Body")
                self.inside_if = False
                body = self.stmtlist()
                else_stmt = ElseStmt( body , l, c, self.debug)
                if not ifstmt.body :
                    ifstmt.body = else_stmt
                else:
                    ifstmt.append_stmt( else_stmt )
                break
            else:
                self.inside_if = False
                exception_msg = get_message(Messages.CaseSyntaxError)
                raise ParseException(exception_msg)
            ptok = self.peek()
            self.dbg_msg("parsing SWITCH-CASE next bits "+str(ptok))
        self.match( tamilToken.END )
        self.inside_if = False
        self.dbg_msg("parsing -SWITCH-CASE- complete")
        return ifstmt
    
    def parseIfStmt(self,exp):
        ## @ <expression> if { stmtlist } @<expr> ELSEIF {stmtlist} ELSE <stmtlist> END
        self.dbg_msg(u"parsing IF statement")
        if_tok = self.dequeue()
        [l,c]=if_tok.get_line_col()
        self.inside_if = True
        ifstmt = IfStmt( exp[0], None, None, l, c, self.debug)
        self.if_stack.append(ifstmt)
        self.dbg_msg(u"parsing IF-body")
        body = self.stmtlist()
        prev_body = body;
        ifstmt.set_body( body )
        ptok = self.peek()
        while ( ptok.kind == tamilToken.ATRATEOF or ptok.kind == tamilToken.ELSE ):
            self.inside_if = True        
            [l,c]=ptok.get_line_col()
            if ( ptok.kind == tamilToken.ATRATEOF ):
                # parse elseif branch
                self.dbg_msg(u"parsing ELSE-IF")                
                self.match( tamilToken.ATRATEOF )
                exp = self.valuelist();
                self.dbg_msg(u"parsing ELSE-IF EXPR")
                tok = self.peek()
                if ( tok.kind != tamilToken.ELSEIF ):
                    # maybe another IF statement, SWITCH-CASE or a WHILE loop, DO-WHILE loop etc.
                    next_stmt = self.stmtlist(exp) #pass in the expression                                        
                    prev_body.append( next_stmt )
                    # append to previously scanned body.
                else:
                    self.dbg_msg(u"parsing ELSE-IF-body")
                    self.match( tamilToken.ELSEIF )
                    body = self.stmtlist()
                    prev_body = body
                    next_stmt = IfStmt(exp[0],body,None,l,c,self.debug)
                    self.dbg_msg(u"ELSEIF parsed correctly => "+str(next_stmt))
                    ifstmt.append_stmt( next_stmt )
            elif ( ptok.kind == tamilToken.ELSE ):
                #parse else branch                
                self.dbg_msg(u"parsing stmt else: ")
                self.match( tamilToken.ELSE )
                self.dbg_msg(u"parsing ELSE-Body")
                self.inside_if = False
                body = self.stmtlist()
                prev_body = body;
                else_stmt = ElseStmt( body , l, c, self.debug)                
                ifstmt.append_stmt( else_stmt )
                break
            else:
                self.inside_if = False
                exception_msg = get_message(Messages.IfSyntaxError)
                raise ParseException(exception_msg)
            ptok = self.peek()
            self.dbg_msg(u"parsing -IF next bits "+str(ptok))
        self.match( tamilToken.END )
        self.inside_if = False
        self.dbg_msg(u"parsing -IF-complete")
        return ifstmt
    
    def stmt(self,pass_in_ATexpr=None):
        """ try an assign, print, return, if or eval statement """
        self.dbg_msg(u" STMT ")
        ptok = self.peek()
        self.dbg_msg(u"stmt: peeking at "+str(ptok))
        if ( ptok.kind ==  tamilToken.RETURN ):
            ## return <expression>
            self.dbg_msg('enter->return: <expression>')
            ret_tok = self.dequeue()
            [l,c]=ret_tok.get_line_col();
            if ( not self.parsing_function ):
                raise ParseException( u"return statement outside of function body "+str(ret_tok))
            rstmt = ReturnStmt(self.expr(),l,c,self.debug)
            self.dbg_msg(u"return statement parsed")
            return rstmt
        elif ( ptok.kind ==  tamilToken.PRINT ):
            self.dbg_msg(u"stmt : print ")
            self.currently_parsing.append( ptok )
            ## print <expression>
            print_tok = self.dequeue()
            [l,c]=print_tok.get_line_col();
            exprlist_val = self.exprlist();
            self.currently_parsing.pop()
            return PrintStmt(exprlist_val,l,c,self.debug)        
        elif ( ptok.kind ==  tamilToken.ATRATEOF or pass_in_ATexpr):
            ## @ <expression> {if | while | elseif}
            if not pass_in_ATexpr:
                at_tok = self.match(tamilToken.ATRATEOF)
                self.currently_parsing.append( at_tok )
                exp = self.valuelist();
                self.currently_parsing.pop()
            else:
                exp = pass_in_ATexpr
                pass_in_ATexpr = None #use it just once
            if( self.debug ): print ("return from valuelist ",str(exp))
            ptok = self.peek();
            if ( ptok.kind == tamilToken.IF ):                
                return self.parseIfStmt(exp)
            elif ( ptok.kind ==  tamilToken.WHILE ):
                ## @ ( expr ) while { body } end               
               self.loop_stack.append(True)
               self.dbg_msg(u"while-statement")
               while_tok = self.dequeue()
               self.currently_parsing.append( while_tok )
               [l,c]=while_tok.get_line_col()
               wexpr = exp[0];
               body = self.stmtlist( )
               self.match( tamilToken.END)
               whilestmt = WhileStmt(wexpr, body, l, c, self.debug)
               self.loop_stack.pop()
               self.currently_parsing.pop()
               return whilestmt
            elif ( ptok.kind ==  tamilToken.SWITCH ):
                return self.parseSwitchStmt(exp)
            elif ( ptok.kind ==  tamilToken.FOREACH ):
                foreach_tok = self.dequeue()
                self.currently_parsing.append(foreach_tok)
                [l,c]=foreach_tok.get_line_col()
                if ( self.debug ): print(u"parsing FOREACH stmt")
                self.loop_stack.append(True)
                self.dbg_msg(u"foreach-statement")
                # convert to a for statement - building tamil AST - transformations
                if not isinstance( exp[1], Identifier ):
                    raise ParseException(u" FOR-EACH statement "+str(foreach_tok) )
                foreach_iter = exp[1];
                iter = Identifier("__"+foreach_iter.id,l=0,c=-1);
                eq_token = tamilLexeme(u"=",tamilToken.EQUALS)
                plus_token = tamilLexeme(u"+",tamilToken.PLUS)
                lt_token = tamilLexeme(u"<",tamilToken.LT)
                if ( self.debug ): print(u"build init assign stmt")
                init = AssignStmt( iter, eq_token , Number(0),l,c,self.debug)
                if ( self.debug ): print(u"build cond expr")
                VL1 = ValueList([exp[0]],l,c,self.debug)
                cond = Expr( iter, lt_token, ExprCall( Identifier("len",l,c), VL1, l, c, self.debug ), l, c, self.debug )
                if ( self.debug ): print("build plus1 stmt")
                plus1_iter = Expr( iter, plus_token, Number(1), l, c, self.debug  )
                if ( self.debug ): print(u"build equals stmt")
                update = AssignStmt( iter, eq_token , plus1_iter ,l,c,self.debug)
                body = self.stmtlist() #parse body
                # and insert artifical update variable in body
                VL2 = ValueList([exp[0],iter],l,c,self.debug)
                extract_foreach_iter_from_list = ExprCall( Identifier("__getitem__",l,c), VL2,l,c,self.debug);
                foreach_iter_Assign = AssignStmt( foreach_iter, eq_token , extract_foreach_iter_from_list, l,c,self.debug )
                body.List.insert( 0,foreach_iter_Assign)
                # complete FOREACH stmt
                self.match( tamilToken.END)
                self.currently_parsing.pop()
                foreach_stmt = ForStmt(init, cond, update, body, l, c, self.debug);                
                self.loop_stack.pop();                
                if ( self.debug ): print(u"completed parsing FOR-EACH loop",str(foreach_stmt))
                return foreach_stmt
            elif ( ptok.kind ==  tamilToken.FOR ):
                ## Fixme : empty for loops not allowed.
                """ For ( exp1 , exp2 , exp3 ) stmtlist  end"""
                if ( self.debug ): print("parsing FOR stmt")
                self.loop_stack.append(True)
                self.dbg_msg("for-statement")                
                for_tok = self.peek()
                self.currently_parsing.append(for_tok)
                if ( self.debug ): print("matching for STMT",str(self.peek()))
                self.match( tamilToken.FOR )
                if ( self.debug ): print("matched for STMT",str(self.peek()))
                [l,c]= for_tok.get_line_col();
                init,cond,update = exp[0],exp[1],exp[2]            
                if ( self.debug ): print("extract 3 parts",str(init),str(cond),str(update))
                body = self.stmtlist()
                self.match( tamilToken.END)
                self.currently_parsing.pop()
                if ( self.debug ): print("body of loop",str(body))
                forstmt = ForStmt(init, cond, update, body, l, c, self.debug);
                self.loop_stack.pop();
                if ( self.debug ): print("completed parsing FOR loop",str(forstmt))
                return forstmt
        elif ( ptok.kind == tamilToken.DO ):
            if ( self.debug ): print("parsing DO-WHILE statement")
            self.loop_stack.append(True)
            do_tok = self.dequeue()
            self.currently_parsing.append(do_tok)
            [l,c]=do_tok.get_line_col()
            body = self.stmtlist()
            if ( self.debug ): print("parsed body")
            self.match(tamilToken.DOWHILE)            
            self.match(tamilToken.ATRATEOF)
            exp = self.valuelist();
            if ( self.debug ): print("parsed EXP",exp[0])
            doWhileStmt = DoWhileStmt(exp[0], body, l, c, self.debug)
            self.loop_stack.pop()
            self.currently_parsing.pop()
            return doWhileStmt
        elif ( ptok.kind ==  tamilToken.BREAK ):
            ## break, must be in loop-environment
            self.dbg_msg("break-statement");
            break_tok = self.dequeue();
            [l,c]=break_tok.get_line_col()
            self.check_loop_stack(); ##raises a parse error
            brkstmt = BreakStmt( l, c, self.debug);
            return brkstmt
        elif ( ptok.kind ==  tamilToken.CONTINUE ):
            ## continue, must be in loop-environment
            self.dbg_msg("continue-statement");
            cont_tok = self.dequeue();
            [l,c]=cont_tok.get_line_col()
            self.check_loop_stack(); ##raises a parse error
            cntstmt = ContinueStmt( l, c, self.debug);
            return cntstmt
        elif ( ptok.kind == tamilToken.IMPORT ):
            self.dbg_msg("import-statement")
            import_tok = self.dequeue()
            [l,c] = import_tok.get_line_col()
            self.currently_parsing.append(import_tok)
            fname = self.expr()
            self.currently_parsing.pop()
            importstmt = ImportStmt(l,c,self.debug,fname)
            return importstmt
        else:
            ## lval := rval
            ptok = self.peek()
            self.currently_parsing.append(ptok)
            [l,c] = ptok.get_line_col()
            lhs = self.expr()
            self.dbg_msg("parsing expr: "+str(lhs))
            ptok = self.peek()
            if ( ptok.kind in  tamilToken.ASSIGNOP ):
                assign_tok = self.dequeue()
                rhs = self.expr()
                [l,c]=assign_tok.get_line_col()
                self.currently_parsing.pop()
                if isinstance(lhs,ExprCall):
                    # however array assignment is carried out through setitem
                    # print 'expr-->',lhs,'rhs-->',rhs
                    newlhs = lhs
                    newlhs.fname = '__setitem__'
                    newlhs.arglist.append(rhs)
                    return EvalStmt(newlhs,l,c,self.debug)
                return AssignStmt( lhs, assign_tok, rhs, l, c, self.debug)
            self.currently_parsing.pop()
            return EvalStmt( lhs, l, c, self.debug )
        raise ParseException("parsing Statement, unknown operators" + str(ptok))
    
    def function(self):
        """ def[kw] fname[id] (arglist) {body} end[kw] """
        if ( self.parsing_function ):
            self.parsing_function = False
            raise ParseException(u" Nested functions not allowed! "+str(self.last_token()))

        self.parsing_function = True
        def_tok = self.dequeue()
        if ( def_tok.kind !=  tamilToken.DEF ):
            raise ParseException(u"unmatched 'def'  in function " +str(def_tok))
        
        id_tok = self.dequeue()
        if ( id_tok.kind !=  tamilToken.ID ):
            raise ParseException(u"expected identifier in function"+str(id_tok))
        
        arglist = self.arglist()
        self.dbg_msg( u"finished parsing arglist" )
        body = self.stmtlist()

        self.match(  tamilToken.END )
        [l,c] = def_tok.get_line_col()
        fval = Function( id_tok.val, arglist, body, l, c, self.debug )
        self.parsing_function = False
        self.dbg_msg( u"finished parsing function" ) 
        return fval

    def valuelist(self):
        """parse: ( expr_1 , expr_2, ... ) """
        valueList = list()
        self.dbg_msg(u"valuelist: ")
        lparen_tok = self.match(  tamilToken.LPAREN )
        while ( self.peek().kind !=  tamilToken.RPAREN ):            
            val = self.expr()
            if ( self.debug ): print(u"val = ",str(val))
            ptok = self.peek()
            if ( self.debug ) : print(u"ptok = ",str(ptok),str(ptok.kind),str(tamilToken.ASSIGNOP))
            if ( ptok.kind in  tamilToken.ASSIGNOP ):
                assign_tok = self.dequeue()
                rhs = self.expr()
                [l,c]=assign_tok.get_line_col()
                lhs = val
                val =  AssignStmt( lhs, assign_tok, rhs, l, c, self.debug)
                if ( self.debug ): print(u"AssignStmt = ",str(val))
                ptok = self.peek()
            else:
                if ( self.debug ): print(u"No-Assign // Expr =",str(val))
            self.dbg_msg(u"valuelist-expr: "+str(val))
            valueList.append( val )
            if  ( ptok.kind ==  tamilToken.RPAREN ):
                break
            elif ( ptok.kind ==  tamilToken.COMMA ):
                self.match(  tamilToken.COMMA )
            else:
                raise ParseException(u" function call argument list "+str(ptok))
        self.match(  tamilToken.RPAREN )
        [l,c] = lparen_tok.get_line_col()
        return ValueList(valueList, l, c, self.debug )

    def arglist(self):
        """parse: ( arg_1, arg_2, ... ) """
        self.dbg_msg( u" ARGLIST " )
        args = list()
        lparen_tok = self.match(  tamilToken.LPAREN )
        while ( self.peek().kind !=  tamilToken.RPAREN ):
            arg_name = self.dequeue()
            args.append( arg_name.val )
            ptok = self.peek()
            if  ( ptok.kind ==  tamilToken.RPAREN ):
                break
            elif ( ptok.kind ==  tamilToken.COMMA ):
                self.match(  tamilToken.COMMA )
            else:
                raise ParseException(u" function definition argument list "
                                     +str(ptok))
        self.match(  tamilToken.RPAREN )
        [l,c] = lparen_tok.get_line_col()
        return ArgList(args , l, c, self.debug )
        
    def exprlist(self):
        """   EXPRLIST : EXPR, EXPRLIST        
        ##  EXPRLIST : EXPR """
        self.dbg_msg( u" EXPRLIST " )
        exprs=[]
        comma_tok = None
        l = 0; c = 0
        while ( not self.lex.end_of_tokens() ):
            exprs.append(self.expr())
            if self.lex.peek().kind !=  tamilToken.COMMA:
                break            
            tok = self.match( tamilToken.COMMA)
            if ( not comma_tok ):
                comma_tok = tok 

        if ( comma_tok ):
            [l,c] = comma_tok.get_line_col()
        self.dbg_msg(u"finished expression list")
        return ExprList(exprs, l, c, self.debug)

    def expr(self):
        self.dbg_msg( u" EXPR " )
        val1=self.term()
        res=val1
        ptok = self.peek()
        if ptok.kind in  tamilToken.ADDSUB:
            binop=self.dequeue()
            if ( ptok.kind == tamilToken.MINUS ):
                val2 = self.term()
            else:
                val2=self.expr()
            [l,c] = binop.get_line_col()
            res=Expr(val1,binop,val2, l, c, self.debug )
        elif ptok.kind ==  tamilToken.LPAREN:
            ## function call
            if ( not isinstance(res, Identifier) ):
                raise ParseException(u"invalid function call"+str(ptok))
            [l,c] = ptok.get_line_col()
            vallist = self.valuelist()
            res=ExprCall( res, vallist, l, c, self.debug )
        
        ptok = self.peek()
        while  ptok.kind in tamilToken.BINOP:
            binop = self.dequeue()
            [l,c] = binop.get_line_col()
            res = Expr( res, binop,self.expr(), l,c,self.debug)
            ptok = self.peek()
        return res
    
    def term(self):
        """ this is a grammar abstraction; 
        but AST only has Expr elements"""
        self.dbg_msg( "term" )
        val1=self.factor()
        res=val1

        tok = self.peek()
        if ( tok.kind in  tamilToken.MULDIV 
             or  tok.kind in  tamilToken.COMPARE 
             or tok.kind in  tamilToken.EXPMOD 
             or tok.kind in tamilToken.BITWISE_AND_LOGICAL ):
            binop=self.dequeue()
            val2=self.term()
            [l,c] = binop.get_line_col()
            res=Expr(val1,binop,val2, l, c, self.debug)
            
        return res
    
    def factor(self):
        self.dbg_msg( "factor" )
        tok=self.peek()
        if tok.kind ==  tamilToken.LPAREN:
            lparen_tok = self.dequeue()
            val=self.expr()
            if self.dequeue().kind!= tamilToken.RPAREN:
                raise SyntaxError("Missing Parens "+str(self.last_token()))
        elif tok.kind ==  tamilToken.NUMBER:
            tok_num = self.dequeue()
            [l, c] = tok_num.get_line_col()
            val = Number( tok.val , l, c, self.debug )
        elif tok.kind == tamilToken.LOGICAL_NOT:
            tok_not = self.dequeue()
            [l, c] = tok_not.get_line_col()
            val = UnaryExpr( self.expr(), tok_not , l, c, self.debug )
            self.dbg_msg("completed parsing unary logical-not expression"+str(val))
        elif tok.kind == tamilToken.BITWISE_COMPLEMENT:
            tok_compl = self.dequeue()
            [l, c] = tok_compl.get_line_col()
            val = UnaryExpr( self.expr(), tok_compl , l, c, self.debug )
            self.dbg_msg("completed parsing unary bitwise-complement expression"+str(val))
        elif tok.kind ==  tamilToken.ID:
            tok_id = self.dequeue()
            [l, c] = tok_id.get_line_col()
            val = Identifier( tok.val , l, c, self.debug )
            ptok = self.peek()
            self.dbg_msg(u"factor: "+str(ptok) + u" / "+ str(tok) )
            if ( ptok.kind ==  tamilToken.LPAREN ):
                ## function call
                [l, c] = ptok.get_line_col()
                vallist = self.valuelist()
                val=ExprCall( val, vallist, l, c, self.debug )
            elif ( ptok.kind ==  tamilToken.LSQRBRACE ):
                ## indexing a array type variable or ID                
                [l,c] = ptok.get_line_col()
                ## replace with a call to __getitem__
                exp = self.factor();
                if ( hasattr(exp,'__getitem__') ):
                    VL2 = ValueList([val,exp[0]],l,c,self.debug)
                else:
                    # when exp is a expression
                    VL2 = ValueList([val,exp],l,c,self.debug)
                val = ExprCall( Identifier("__getitem__",l,c), VL2,l,c,self.debug)
                for itr in range(1,len(exp)):
                    VL2 = ValueList([val,exp[itr]],l,c,self.debug)
                    val = ExprCall( Identifier("__getitem__",l,c), VL2,l,c,self.debug)
            elif ( ptok.kind ==  tamilToken.LCURLBRACE ):
                val=None
                raise ParseException("dictionary indexing uses square brackets '[' only ']'"+str(ptok));
        elif tok.kind ==  tamilToken.STRING :
            str_tok = self.dequeue()
            [l,c] = str_tok.get_line_col()
            val = String( tok.val , l, c, self.debug )
        elif tok.kind in tamilToken.ADDSUB:
            unop = self.dequeue();
            [l, c] = unop.get_line_col()
            val=Expr(Number(0),unop,self.term(),l,c,self.debug);
        elif tok.kind == tamilToken.LCURLBRACE:
            # creating a list/dictionary expression
            dict_start = self.dequeue();
            val = Dict()
            while( True ):
                if ( self.peek().kind == tamilToken.RCURLBRACE ):
                    break;
                exprkey = self.expr()
                tok_colon = self.match(tamilToken.COLON)
                exprval = self.expr()
                val.update( {exprkey : exprval}  )
                if self.debug : print(self.peek().__class__,self.peek())
                if ( self.peek().kind == tamilToken.RCURLBRACE ):
                    break
                else:
                    assert( self.peek().kind == tamilToken.COMMA)
                    self.dequeue()
            assert( self.peek().kind == tamilToken.RCURLBRACE )
            list_end = self.dequeue()
        elif tok.kind == tamilToken.LSQRBRACE:
            # creating a list/array expression
            list_start = self.dequeue();
            val = Array()
            while( True ):
                if ( self.peek().kind == tamilToken.RSQRBRACE ):
                    break;
                exprval = self.expr()
                val.append( exprval  )
                if self.debug : print(self.peek().__class__,self.peek())
                if ( self.peek().kind == tamilToken.RSQRBRACE ):
                    break
                else:
                    assert( self.peek().kind == tamilToken.COMMA)
                    self.dequeue()
            assert( self.peek().kind == tamilToken.RSQRBRACE )
            list_end = self.dequeue()
        else:
            exception_msg = get_message(Messages.UnexpectedNumber,str(tok))
            raise ParseException(exception_msg)
        self.dbg_msg( u"factor-returning: "+str(val) )
        return val



#!/usr/bin/python
# -*- coding: utf-8 -*-
## 
## (C) 2015 Muthiah Annamalai
## Licensed under GPL Version 3
## 
## Interpreter for tamil language
from __future__ import print_function

import sys
PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

## Tx
from .transform import Visitor, TransformVisitor
from .tamil_scanner import tamilToken
## AST elements
from .ast import Expr, ExprCall, ExprList, Stmt, ReturnStmt, \
 BreakStmt, ContinueStmt, ElseStmt, IfStmt, WhileStmt, \
 ForStmt, AssignStmt, PrintStmt, EvalStmt, ArgList, \
 ValueList, Function, StmtList, Identifier, Number, \
 String, Boolean

from .errors import RuntimeException, SemanticException

class TransformEntryExitProfile(TransformVisitor):
    def __init__(self,**kwargs):
        TransformVisitor.__init__(self,**kwargs)
        
    def visit_program_or_script(self,stmt_list):
        l,c=0,0
        stmt_list.dbg_msg(" add call : profile(\"begin\")")
        begin = ValueList([String("begin")],l,c,self.debug)
        call_profile_begin = ExprCall( Identifier("profile",l,c), begin, l, c, self.debug )
        stmt_list.List.insert(0,call_profile_begin)
        
        stmt_list.dbg_msg(" add call : 'profile(\"results\")'")
        results = ValueList([String("results")],l,c,self.debug)
        call_profile_results = ExprCall( Identifier("profile",l,c), results, l, c, self.debug )
        stmt_list.append( call_profile_results )
        return

class TransformSafeModeFunctionCheck(TransformVisitor):
    def __init__(self,**kwargs):
        self.forbidden_fcn_names = [u'raw_input',u'input',u'fopen',u'open',u'fclose',\
        u'உள்ளீடு',u'turtle',u'கோப்பை_எழுது',u'கோப்பை_திற',u'கோப்பை_மூடு']
        TransformVisitor.__init__(self,**kwargs)
        
    def visit_expr_call(self,expr_call):
        callee = expr_call.func_id.id
        if callee in self.forbidden_fcn_names:
           raise RuntimeException(u"ERROR %s:\n\t %s may not be used in SAFE MODE ."%(self.interpreter.get_fname(),str(expr_call)))
        if expr_call.arglist:
            expr_call.arglist.visit( self )
        return

#  Type checker for tamil - rules list #65
class TransformSemanticAnalyzer(TransformVisitor):
    def __init__(self,**kwargs):
        TransformVisitor.__init__(self,**kwargs)
        return
    
# Find a list of rules for type checking tamil AST.
# You may only add like types. I.e. (You may only add numbers or strings but never between each other)
# You may index arrays with only integers or numbers or dictionaries with Strings
# You can type check argument types, and number of builtin functions.
# You may type check arguments for number of args in a function call.
    
    def visit_expr_call(self,expr_call):
        callee = expr_call.func_id.id
        if callee == u"__getitem__":
            # T.B.D
            pass
        if expr_call.arglist:
            expr_call.arglist.visit( self )
        return
    
    # check if the constants are on lhs of assignment statements
    # check if the strings are added to numbers 
    # check ...
    def visit_assign_stmt(self, assign_stmt):
        if any( map( lambda typename: isinstance( assign_stmt.lvalue, typename) , [Number,String,Boolean,Function]) ):
            raise SemanticException("Cannot use number, string, constant or functions on LHS of assignment %s"%str(assign_stmt))
        if assign_stmt.lvalue:
            assign_stmt.lvalue.visit( self )
        if assign_stmt.rvalue:
            assign_stmt.rvalue.visit( self )
        return
    
    def visit_binary_expr(self,binexpr):
        lhs_is_string = isinstance( binexpr.term, String)
        rhs_is_string = isinstance( binexpr.next_expr, String )
        lhs_id_expr_call = isinstance( binexpr.term, ExprCall ) or isinstance( binexpr.term, Identifier)
        rhs_id_expr_call = isinstance( binexpr.next_expr, ExprCall ) or isinstance( binexpr.next_expr, Identifier)
        
        if isinstance(binexpr.next_expr,Expr):
            binexpr.next_expr.visit(self)
            return
        
        binexpr.term.visit(self)
        
        if binexpr.binop.kind != tamilToken.PLUS:
            if lhs_is_string or rhs_is_string:
                if binexpr.binop.kind in tamilToken.COMPARE or binexpr.binop.kind == tamilToken.PROD:
                    pass
                else:
                    raise SemanticException("Cannot use string with operators other than '+','>=','<=','!=','==','>','<' or '*' at expression %s %s"%(str(binexpr),binexpr.get_pos()))
        else:
            if lhs_is_string or rhs_is_string:
                if not ((lhs_is_string and rhs_is_string) or \
                   (lhs_is_string and rhs_id_expr_call) or \
                   (rhs_is_string and lhs_id_expr_call)):
                   raise SemanticException("Cannot join strings and expression at expression %s"%str(binexpr))
        return
    
    def visit_import(self,importstmt):
        if not isinstance(importstmt.filename,String):
            raise SemanticException("Import statement should be a string at time of interpretation at %s"%str(importstmt))
        return

class TransformConstantFolder(TransformVisitor):
    def __init__(self,**kwargs):
        TransformVisitor.__init__(self,**kwargs)
        self.rval = None
        #print(self.top_ast)
        return
        
    def constant_fold(self,binexpr):
        return binexpr.evaluate(None)
    
    def can_fold_expr(self,expr):
        if isinstance(expr,Number):
            return True, expr
    
    def reset(self):
        self.rval = None
    
    def get_rval(self):
        op = self.rval
        self.reset()
        return op
    
    def visit_number(self,num):
        self.rval = num
        return
        
    def visit_binary_expr(self,binexpr):
        # if lhs is constant and you are able to fold rhs
        # then replace binexpr with the value
        #print(type(binexpr.term))
        #print(type(binexpr.next_expr))
        next_expr_alt = None
        if isinstance(binexpr.next_expr,Expr):
            binexpr.next_expr.visit(self)
            next_expr_alt = self.get_rval()
        else:
            next_expr_alt = binexpr.next_expr
        
        binexpr.term.visit(self)
        term_expr_alt = self.get_rval()
        
        print(type(term_expr_alt))
        #print("-------")
        if next_expr_alt == None or term_expr_alt == None:
            return None
        #print("------x------")
        
        lhs_is_num = isinstance( term_expr_alt, Number)
        [foldable,val] = self.can_fold_expr( next_expr_alt )
        
        if foldable:
            print("foldable")
            # new API needed to replace the node
            binexpr.term = term_expr_alt
            binexpr.next_expr = next_expr_alt
            newval = self.constant_fold( binexpr )
            binexpr.replace( newval )
            print(str(newval),newval)
            return Number(newval)
        return None
    


#!/usr/bin/python
# -*- coding: utf-8 -*-
## 
## (C) 2013 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## tamil language Interpreter via Web

## Ref: http://wiki.python.org/moin/ http.serverfrom __future__ import print_function
import time
import os
from tamil import tamilFileExecuter, tamilInterpExecuter
import http.server, threading
from http.server import SimpleHTTPRequestHandler
from socketserver import ThreadingMixIn
from os import unlink
from urllib.parse import parse_qs
import urllib.parse, codecs
import sys, traceback

DEBUG=False

class BasetamilWeb(SimpleHTTPRequestHandler):    
    def do_GET(self):
        print(str(self.headers), "in thread =", threading.currentThread().getName())    
        #delegate to parent
        SimpleHTTPRequestHandler.do_GET(self)
        return
    
    def do_POST(self):
        query_string = self.rfile.read(int(self.headers['Content-Length']))
        print(query_string)
        POSTvars = urllib.parse.parse_qs( query_string )
        print(str(POSTvars))
        
        if POSTvars.has_key('prog'):
            program ="\n".join(POSTvars['prog'])
            program = codecs.decode( program ,'utf-8')
        elif POSTvars.has_key('eval'):
            program = 'printf("Welcome to tamil! You can type a program and execute it online!")'
        else:
            self.send_response(404)
            self.send_header("Content-type","text/html")
            self.end_headers()
            self.wfile.write( "<h1>HTTP 404 : Error occured</h1>" )
            self.wfile.write( str(POSTvars) )
            return
        
        self.send_response(200)
        self.send_header("Content-type","text/html; charset=utf-8")
        self.end_headers()
        self.do_tamil_execute( program )        
        
        return
    
    def do_tamil_execute(self,program):
        # write the input program into a temporary file and execute the tamil Interpreter
        
        program_fmt = u"""<TABLE>
        <TR><TD>
        <TABLE>
        <TR>
        <TD><font color=\"blue\"><OL>"""
        
        print( u"Source program" )
        print( program )
        print( u"*"*60 )
        
        program_fmt += u"\n".join(["<li>%s</li>"%(prog_line)  for line_no,prog_line in enumerate(program.split('\n'))])
        program_fmt += u"""</OL></font></TD></TR>\n</TABLE></TD><TD>"""
        
        # run the interpreter in a sandbox and capture the output hopefully
        try:
            failed = True #default failed mode
            obj = tamilFileExecuter( file_input = [program], redirectop = True, TIMEOUT = 60*2 ) # 2 minutes
            
            # actually run the process
            obj.run()
        
            # get executed output in 'progout' and name of the two tmp files to cleanup
            [tmpfile,filename,progout] = obj.get_output()
            
            for f in [tmpfile,filename]:
                try:
                    os.unlink( f )
                except Exception as e:
                    pass
                
            if obj.exitcode != 0: #and tamilWeb.error_qualifiers(progout)
                failed = True
            else:
                failed = False

            # output from tamil interpreter is in form of UTF-8 strings, we extract
            # it from ASCII format I/O since we are in CGI mode.
            progout = progout.decode('utf-8')
            if DEBUG:
                print(u"output = ")
                print(progout)
            
            #SUCCESS_STRING = "<H2> Your program executed correctly! Congratulations. </H2>"
            FAILED_STRING = "Traceback (most recent call last)"
            if obj.exitcode != 0 and progout.find(FAILED_STRING) > -1:
                print(u"Exitcode => ",obj.exitcode)
                op = u"%s <B>FAILED Execution, with parsing or evaluation error</B> for program with <font color=\"red\">error <pre>%s</pre> </font></TD></TR></TABLE>"%(program_fmt,progout)
            else:
                failed = False
                obj.exitcode = 0
                op = u"%s <B>Succeeded Execution</B> for program with output, <BR/> <font color=\"green\"><pre>%s</pre></font></TD></TR></TABLE>"%(program_fmt,progout)
        except Exception as e:
            print(u"FAILED EXECUTION",str(e))
            traceback.print_tb(sys.exc_info()[2])
            failed = True
            op = u"%s <B>FAILED Execution</B> for program with <font color=\"red\">error <pre>%s</pre> </font></TD></TR></TABLE>"%(program_fmt,str(e))
        else:
            print(u"Output file")
            obj.get_output()
        
        prev_page = u"""<script>
    document.write("Navigate back to your source program : <a href='#' onClick='history.back();return false;'>Go Back</a>");
</script><HR/>"""
        #op = ""
        if failed:
            op = u"<H2> Your program has some errors! Try correcting it and re-evaluate the code</H2><HR/><BR/>"+op
        else:
            op = u"<H2> Your program executed correctly! Congratulations. </H2><HR/><BR/>"+op            
        op = prev_page + op
        real_op = u"<html> <head> <title>tamil interpreter</title> </head><body> %s </body></html>\n"%op
        
        # CGI pipe only allows ASCII style strings
        self.wfile.write(real_op.encode('utf-8'))
        
        return op

class tamilWeb(ThreadingMixIn,BasetamilWeb):
    """ Add threading to handle requests in separate thread """
    pass

HOST_NAME = "localhost"
PORT_NUMBER = 8080

if __name__ == "__main__":
    httpd =http.server.HTTPServer((HOST_NAME, PORT_NUMBER), tamilWeb)
    print(time.asctime(), "Server Starts - %s:%s" % (HOST_NAME, PORT_NUMBER))
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()
    print(time.asctime(), "Server Stops - %s:%s" % (HOST_NAME, PORT_NUMBER))



from __future__ import print_function
import sys
from .runtime import BuiltinFunction
import json


PYTHON3 = (sys.version[0] == '3')

if PYTHON3:
    # For Python 3.0 and later
    from urllib.request import urlopen
else:
    # Fall back to Python 2's urllib2
    from urllib import urlopen

def tamil_urlopen(*args):
    html = urlopen(*args)
    return html

def tamil_urlread(html):
    data = html.read()
    if not PYTHON3:
        return data.decode("utf-8")
    return data

def tamil_urlclose(html):
    html.close()

def Load_URL_APIs(interpreter):
    interpreter.add_builtin("urlopen",tamil_urlopen,nargin=1,ta_alias="இணைய_இணைப்பு_திற")
    interpreter.add_builtin("urlread",tamil_urlread,nargin=1,ta_alias="இணைய_இணைப்பு_படி")
    interpreter.add_builtin("urlclose",tamil_urlclose,nargin=1,ta_alias="இணைய_இணைப்பு_மூடு")
    # JSON
    interpreter.add_builtin("json_loads",json.loads,nargin=1)
    interpreter.add_builtin("json_load",json.load,nargin=1)
    interpreter.add_builtin("json_dump",json.dump,nargin=1)
    interpreter.add_builtin("json_dumps",json.dumps,nargin=1)



#!/usr/bin/python
# -*- coding: utf-8 -*-
## 
## (C) 2007, 2008, 2013 Muthiah Annamalai,
## Licensed under GPL Version 3
## 
## Module has elements of PARSE-TREE  AST 
## in the classes Expr, ExprCall, ExprList, Stmt, ReturnStmt,
## BreakStmt, ContinueStmt, ElseStmt, IfStmt, WhileStmt, 
## ForStmt, AssignStmt, PrintStmt, EvalStmt, ArgList,
## ValueList, Function, StmtList, Identifier, String, Number,
## Array, Dict, NoOp

import keyword
import copy
import math
import sys

PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str

## scanner for exprs language
from .scanner import Token, Lexeme, Lex

## runtime elements
from .runtime import  Environment, BuiltinFunction, \
 BlindBuiltins

## exceptions
from .errors import RuntimeException, ParseException

##
## ATOMS
##

class Identifier(object):
    def __init__(self,id, l, c, dbg=False):
        object.__init__(self)
        self.id=id
        self.debug = dbg
        self.line = l;
        self.col = c;
    
    def __str__(self):
        return u"" + self.id
    
    def dbg_msg(self, msg):
        """ handy to print debug messages """
        if ( self.debug ):
            print(u"## "+msg)
        return

    def __repr__(self):
        return u"\n\t [Identifier [" + str(self.id) +u"]]"

    def evaluate(self,env):
        if ( env.has_id(self.id)):
            val = env.get_id(self.id)
            if ( hasattr( val, 'evaluate' ) ):
                val = val.evaluate(env)
            elif ( val.__class__ == str ):
                #val = val
                pass
            else:
                #val = val
                pass
            self.dbg_msg( str(self) + " = val ["+str(val) + "]" )
            return val
        note = ''
        if self.id in keyword.kwlist:
            note = 'Did you possibly confuse the Python english keyword %s for tamil keyword ?'%self.id
        note = "Cannot Find Identifier %s at Line %d, col %d."%( self.id, self.line,self.col ) + ' ' + note
        raise RuntimeException(note)

    def visit(self, walker):
        """ visitor - do something with a identifier """
        walker.visit_identifier(self)
        return

class String(object):
    def __init__(self,s, l = 0, c = -1, dbg = False):
        object.__init__(self)
        self.string = s
        self.debug = dbg
        self.line = l
        self.col = c
    
    def __repr__(self):
        return u" [String [" + str(self.string) + u"]] "
    
    def __str__(self):
        return self.string

    def __str__(self):
        return ( self.string )

    def evaluate(self,env):
        return self.string

    def visit(self, walker):
        walker.visit_string(self)
        return

class Number(object):
    def __init__(self,n, l = 0, c = -1, dbg = False):
        object.__init__(self)
        self.num=n
        self.debug = dbg
        self.line = l
        self.col = c

    def __int__(self):
        return int(self.num)

    def __float__(self):
        return float(self.num)
    
    def __repr__(self):
        return u" [Number [" + str(self.num) + u"]]"
    
    def __str__(self):
        return self.num.__str__()

    def __str__(self):
        return str( self.num )

    def evaluate(self,env):
        return self.num

    def visit(self, walker):
        walker.visit_number(self)
        return

class Boolean(Number):
    def __init__(self,n, l = 0, c = -1, dbg = False):
        Number.__init__(self,n,l,c,dbg)
    
    def __str(self):
        return self.__str__()
    
    def __str__(self):
        if ( self.num ):
            return u"மெய் (T)"
        return u"பொய் (F)"

class Dict(dict):
    def __init__(self):
        dict.__init__(self)
        
    def base_evaluate(self,env):
        rval = {}
        for x,y in self.items():
            rval.update({ x.evaluate( env ): y.evaluate( env )} )
        return rval
    
    def __str__(self):
        fmt = u"{"
        for k,v in self.items():
            fmt = fmt + str(k) + u" : " + str(v) + u",\n"
        fmt = fmt + u"}"
        return fmt

    def visit(self,walker):
        return walker.visit_dict(self)
    
    def evaluate(self,env):
        """ how do you evaluate dictionaries? just return the favor """
        return self.base_evaluate( env )

class Array(list):
    def __init__(self):
        pass

    def base_evaluate(self,env):
        rval = []
        for v in self:
            rval.append( v.evaluate(env) )
        return rval
    
    def __str__(self):
        return u", ".join( [str(item) for item in self] )
    
    def visit(self,walker):
        return walker.visit_array(self)
    
    def evaluate(self,env):
        return self.base_evaluate( env )

class ExprCall(object):
    """handle function call statement etc."""
    def __init__(self,func_id,arglist, l, c, dbg = False):
        object.__init__(self)
        self.func_id = func_id 
        self.fname = func_id.id
        self.arglist = arglist
        self.debug = dbg
        self.line = l
        self.col = c
        self.parent = None

    def dbg_msg(self, msg):
        if ( self.debug ):
            print("## ",msg)
        return
    
    def __str__(self):
        return u"Line %d, Column %d : Function call [%s] with [%d] args"%(self.line,self.col,str(self.fname),len(self.arglist))
    
    def __repr__(self):
        return u"\n\t [ExprCall[ "+str(self.fname)+u" (" \
            +str(self.arglist)+u")]]"
    
    def evaluate(self,env):
        #self.dbg_msg( str(env) )
        if ( self.debug ):
            print(u"\n".join( env.builtin_map.keys() ))
            print("*"*60)
            print(u"\t".join(env.function_map.keys() ))
            print( self.fname, " ==?== ",env.builtin_map.get(self.fname,None))
        if ( env.has_function(self.fname) ):
            self.dbg_msg("calling function "+ self.fname+ " with %d args"%len(self.arglist))
            fval = env.get_function(self.fname)
            #("Check arguments accepted by function against supplied by call site")
            if hasattr(fval,'arglist'):
                # only check the argument matching for custom functions
                # 'BlindBuiltins' or 'BlindFunctions' donot get argument checked.
                expected_args = len(fval.arglist)
                actual_args = len(self.arglist)
                if expected_args != actual_args:
                    raise RuntimeException(u"Function '%s' expects %d arguments but received only %d at site:\n\t %s\n"%(self.fname,expected_args,actual_args,str(self)))
            ## use applicative order evaluation.
            eval_arglist = [ i.evaluate(env) for i in self.arglist.get_list()];
            env.set_args(  eval_arglist )
            try:
                rval = fval.evaluate(env)
            except Exception as e:
                raise RuntimeException( str(e) )
            self.dbg_msg( u"function retval ="+str(rval)+str(type(rval)))
        else:
            raise RuntimeException(u"undefined function: %s near ( %d, %d )"%(self.fname, self.line, self.col) )
        return rval

    def visit(self,walker):
        walker.visit_expr_call(self)
        return
    
class ExprList(object):
    def __init__(self,exprs, l, c, dbg=False):
        object.__init__(self)
        self.exprs = exprs
        self.debug = dbg
        self.line = l
        self.col = c
        self.parent = None
    
    def __repr__(self):
        return u"\n\t [ExprList[ "+ u", ".join(map(str,self.exprs)) + u"]]"

    def evaluate(self,env):
        """evaluate  a, b, c ... z to a string w/o commas"""
        z = []
        for exp_itr in self.exprs:
            z.append(exp_itr.evaluate(env))
        return u" ".join(map(str,z))
    
    def visit(self, walker):
        walker.visit_expr_list(self)
        return    

class Stmt(object):
    def __init__(self, l=0, c=0, dbg =False):
        """ implements an empty statement"""
        object.__init__(self)
        self.line = l
        self.col = c
        self.class_name = u"Stmt"
        self.debug = dbg
        self.parent = None
        
    def __str__(self):        
        self.dbg_msg( u" ".join([ u"stmt => ", str(self.__class__)]) ) #we're headed toward assertion
        return self.__repr__()

    def __repr__(self):
        print("//#//"*50)
        print(u"stmt => ", str(self.__class__)) #we're headed toward assertion
        self.dbg_msg(u"stmt => "+ str(self.__class__) )
        raise Exception(u"FATAL : Class %s did not implement the __repr__ method, nor inherits a concrete implementation."%str(self.__class__))
    
    def dbg_msg(self, msg):
        """ handy to print debug messages """
        if ( self.debug ):
            print(msg)
        return
    
    def get_pos(self):
        return u"line %d, col %d"%(self.line,self.col)
        
    def evaluate(self, env):
        """ empty statement """
        return None

    def is_true_value(self, val):
        """ Decide if the val is agreeable to True.
        Right now keep it simple however."""
        rval = False
        self.dbg_msg(u"is_true_value? "+ str(val.__class__))
        try:
            if ( hasattr(val,'evaluate') ):
                fval = val.evaluate(None);
            elif ( isinstance(val,float) or isinstance(val,int) ):
                fval = val
            else:
                raise  Exception(u"Unknown case, cannot identify truth @ "+self.get_pos()+u" for value "+str(val))
            
            if ( fval > 0.0 ):
                rval = True
            ## all other cases later.
        except Exception as pyEx:
            """ objects where is_true_value() is not supported """
            print(pyEx)
            raise RuntimeException(pyEx)
        self.dbg_msg(u"Is True Value? " + str(rval) + str(val.__class__) )
        return rval
    
    def visit( self, walker):
        walker.visit_stmt( self )
        return

class DeclarationStmt(Stmt):
    """ hold function declaration statements; have visit option, 
        but no evaluation options. """
    def __init__(self,fcn):        
        if isinstance(fcn,Function):
            Stmt.__init__(self,fcn.line,fcn.col,fcn.debug)
            self.class_name = u"Declaration_Statement"
            self.fcn = fcn #FunctionStmt object
        else:
            raise Exception(u"declaration statement can only hold FunctionStmt objects")
    
    def visit( self, walker):
        """ delegate visitor to holding function """
        walker.visit_function( self.fcn )
        return
    
    def __repr__(self):        
        return self.fcn.__repr__()

class ImportStmt(Stmt):
    """ hold function declaration statements; have visit option, 
        but no evaluation options. """
    def __init__(self,line,col,debug,fname):
        Stmt.__init__(self,line,col,debug)
        self.class_name = u"Import_Statement"
        self.filename = fname
        
    def evaluate(self,env):
        # make a function call to tamil_execute
        #self.dbg_msg(" add call : execute(\"begin\")")
        [l,c] = self.line,self.col
        fname = ValueList([self.filename],l,c,self.debug)
        import_via_execute = ExprCall( Identifier("execute",l,c), fname, l, c, self.debug )
        import_via_execute.evaluate(env)
        return
    
    def visit( self, walker):
        """ delegate visitor to holding function """
        walker.visit_import( self )
        return
    
    def __repr__(self):
        return u"ImportStmt @ %s"%str(self.filename)

class UnaryExpr(Stmt):
    def __init__(self,t,op,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.term=t
        self.unaryop=op
        
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return u"[UnaryExpr["+str(self.unaryop)+ ","+str(self.term)+"]]"
    
    def visit(self,walker):
        """ delegate visitor to transformer/walker"""
        walker.visit_unaryexpr(self)
        return
    
    def do_unaryop(self,tval):
        if ( self.unaryop.kind == Token.LOGICAL_NOT ):
            if not tval:
                return Boolean( True )
            else:
                return Boolean( False )
        elif ( self.unaryop.kind == Token.BITWISE_COMPLEMENT):
            # do the unary bitwise complement
            return ~ tval
        else:
            raise RuntimeException(" unknown Unary Operation - "+str(self.unaryop)+" not supported")
        return
    
    def evaluate(self,env):
        term=self.term.evaluate(env)
        self.dbg_msg(u"unaryop=> "+str(term) +u" "+ str(term.__class__))
        if self.unaryop.kind in Token.UNARYOP:
            tval = Expr.normalize_values( self, term, env)
            if ( self.debug ): print(tval, type(tval))
            term = self.do_unaryop( tval )
        else:
            raise RuntimeException(" unknown Unary Operation - "+str(self.unaryop)+" not supported")        
        self.dbg_msg(u"unaryop=> "+u"term = "+str(term)+u" "+str(term.__class__))
        return term

class Expr(Stmt):
    One = ( 1 );
    Zero = ( 0 );
    def __init__(self,t,op,next_expr,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.term=t
        self.binop=op
        self.next_expr=next_expr        

    def __len__(sef):
        """ expr is unit length always, as opposed to ExprList which is a n-len thing """
        return 1
    
    def __repr__(self):
        return u"\n\t [Expr[ "+ str(self.term)+ u"] " + \
               Token.token_types[self.binop.kind] + \
               u"\t NextExpr [" + str(self.next_expr) + u"]]"

    def do_binop(self,opr1,opr2, binop):
        self.dbg_msg(u" Doing binary operator " + Token.token_types[binop])
        if binop == Token.PLUS:
            self.dbg_msg("addition")
            val = (opr1+opr2)
        elif binop == Token.MINUS:
            self.dbg_msg("subtraction")
            val = (opr1-opr2)
        elif binop == Token.PROD:
            self.dbg_msg("multiplication")
            val = (opr1*opr2)
        elif binop == Token.DIV:
            self.dbg_msg("division")
            val = (opr1/opr2)
        elif binop == Token.MOD:
            self.dbg_msg("modulo")
            val = (math.fmod(opr1,opr2));
        elif binop == Token.EXP:
            self.dbg_msg("exponent")
            val = (math.pow(opr1,opr2));
        elif binop == Token.GT:
            self.dbg_msg("GT > ")
            val = self.Zero;
            if ( opr1 > opr2 ):
                val = self.One;
        elif binop == Token.GTEQ:
            self.dbg_msg("GTEQ >= ")
            val =  self.Zero;
            if ( opr1 >= opr2 ):
                val = self.One;
        elif binop == Token.LT:
            self.dbg_msg("LT < ")
            val =  self.Zero;
            if ( opr1 < opr2 ):
                val = self.One;
        elif binop == Token.LTEQ:
            self.dbg_msg("LT <= ")
            val =  self.Zero;
            if ( opr1 <= opr2 ):
                val = self.One;
        elif binop == Token.NEQ:
            self.dbg_msg("NEQ != ")
            val =  self.Zero;
            if ( opr1 != opr2 ):
                val = self.One;
        elif binop == Token.EQUALITY:
            ## FIXME: do many equality tests than just value
            ## based tests.
            self.dbg_msg("EQUALITY ==  ")
            val =  self.Zero;
            if ( opr1 == opr2 ):
                val = self.One;
        elif binop == Token.LOGICAL_AND:
            self.dbg_msg("LOGICAL AND")
            val = self.Zero
            if ( opr1 and opr2 ):
                val = self.One;
        elif binop == Token.LOGICAL_OR:
            self.dbg_msg("LOGICAL OR")
            val = self.Zero
            if ( opr1 or opr2 ):
                val = self.One;
        elif binop == Token.BITWISE_AND:
            val = opr1 & opr2
        elif binop == Token.BITWISE_OR:
            val = opr1 | opr2
        elif binop == Token.BITWISE_LSHIFT:
            val = opr1 << opr2
        elif binop == Token.BITWISE_RSHIFT:
            val = opr1 >> opr2
        else:
            raise SyntaxError("Binary operator syntax not OK @ "+self.get_pos())
        self.dbg_msg("value = "+str(val))
        return val

    @staticmethod
    def normalize_values( obj, term, env ):
        if ( hasattr(term,'evaluate') ):
            if ( isinstance(term,Number) ): #work for both Number, and derived Boolean class
                tval = term.num
            elif ( isinstance(term,String) ):
                tval = term.string                            
            else:
                raise RuntimeException( " cannot normalize token; unknown clause,"+str(term)+", to evaluate @ "+obj.get_pos());
        else:
            tval = term #float cast not required.
        return tval
    
    def evaluate(self,env):
        term=self.term.evaluate(env)
        self.dbg_msg(u" "+str(term)+u" "+ str(term.__class__))
        if self.binop.kind in Token.BINOP:
            tnext = self.next_expr.evaluate(env)
            tval = Expr.normalize_values( self, term, env)
            tval2 = Expr.normalize_values( self, tnext, env)
            self.dbg_msg(u" "+str( tval)+ " "+str(tval2)+ u" " + str(tval2.__class__))
            try:
                term = self.do_binop(tval, tval2, self.binop.kind)
            except Exception as binOp_Except:
                raise RuntimeException(u"binary operation "+str(self.term)+str(self.binop)+str(self.next_expr)+u" failed with exception "+str(binOp_Except))
        else:
            raise RuntimeException(u" unknown Binary Operation - Binary operation "+str(self.binop)+u" not supported")
        self.dbg_msg(u"term = "+str(term)+u" "+str(term.__class__))
        return term

    def visit(self, walker):
        if self.binop.kind in Token.BINOP:
            walker.visit_binary_expr(self)
        else:
            walker.visit_expr(self)
        return

class ReturnStmt(Stmt):
    """ return expr """
    def __init__(self,rval,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.rvalue=rval
        
    def __repr__(self):
        return u"\n\t [ReturnStmt[ "+ str(self.rvalue)+ u"]]\n"

    def evaluate(self,env):
        rhs=self.rvalue.evaluate(env)
        self.dbg_msg(u"return statement evaluated to "+str(rhs))
        env.set_retval(rhs)
        return rhs
    
    def visit(self, walker):
        walker.visit_return_stmt(self)
        return

class BreakStmt(Stmt):
    """ return expr """
    def __init__(self,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)

    def __repr__(self):
        return u"\n\t [BreakStmt]\n"

    def evaluate(self,env):
        self.dbg_msg(u"break statement")
        env.set_break()
        return None
    
    def visit(self, walker):
        walker.visit_break_stmt(self)
        return

class ContinueStmt(Stmt):
    """ return expr """
    def __init__(self,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
    
    def __repr__(self):
        return u"\n\t [ContinueStmt]\n"

    def evaluate(self,env):
        self.dbg_msg(u"continue statement")
        env.set_continue()
        return None
    
    def visit(self, walker):
        walker.visit_continue_stmt(self)
        return

class ElseStmt(Stmt):
    def __init__(self,stmt,l,c, dbg):
        Stmt.__init__(self,l,c,dbg)
        self.stmt = stmt
        self.class_name = u"ElseStmt"

    def __repr__(self):        
        return u"\t [ElseStmt ["+str(self.stmt) + u"]]\n"
    
    def evaluate(self,env):
        return self.stmt.evaluate(env)
    
    def visit(self,walker):
        walker.visit_else_stmt(self)
        return

class IfStmt(Stmt):
    """ if ( op ) stmtlist {else | elseif ( op )| stmt } end"""
    def __init__(self,expr,body,next_stmt,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.expr = expr
        self.body = body
        self.class_name = "IfStmt"
        ## this is either another IfStmt or an Else Stmt.
        if not next_stmt:
            self.next_stmt = []
        else:
            self.next_stmt = next_stmt
        
    def __repr__(self):
        rval = u"\t\n [IfStmt[["+str(self.expr)+ u"]] "+str(self.body)
        if ( self.next_stmt ):
            try:
                self.dbg_msg( u" ".join([str(self.next_stmt),str(self.next_stmt.__class__),u"***"]) )
                rval = rval + u"<<Nxt>>" + str(self.next_stmt)
            except UnicodeEncodeError as uc_err:
                print(str(uc_err))
                raise uc_err
            pass
        rval = rval + u"]"
        return rval

    def set_body(self,body):
        self.body = body
    
    def append_stmt(self,stmt):
        self.next_stmt.append(stmt)
        return
        
    def set_next_stmt(self, stmt):
        self.next_stmt = stmt
    
    def evaluate(self,env):
        self.dbg_msg( u"Eval-if-stmt" + str(self.expr) )
        rval = None
        self.dbg_msg(u"eval-if stmt")
        if ( self.is_true_value ( self.expr.evaluate(env) ) ):
            self.dbg_msg(u"ifstmt: true condition")
            rval = self.body.evaluate( env )
            return rval
        self.dbg_msg(u"ifstmt: false condition")
        for elseif_or_else in self.next_stmt:
            if ( isinstance( elseif_or_else, IfStmt ) ):
                if ( self.is_true_value( elseif_or_else.expr.evaluate(env) ) ):
                    rval = elseif_or_else.body.evaluate( env )
                    return rval
                else:
                    # elseif branch was found to be false. Continue
                    continue;
            elif( isinstance( elseif_or_else, ElseStmt ) ):
                rval = elseif_or_else.evaluate( env )
                return rval
            else:
                raise RuntimeException("IF-ELSEIF-ELSE was parsed wrongly, unknown construct found")
        # its perfectly legal to not have an else statement
        return rval 
    
    def visit(self,walker):
        walker.visit_if_elseif_stmt(self)
        return

class WhileStmt(Stmt):
    """ while ( exp ) stmtlist  end"""
    def __init__(self,expr,body,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.expr = expr
        self.body = body
        self.class_name = "WhileStmt"

    def __repr__(self):
        rval = u"\t\n [%s[["%str('WhileStmt')+str(self.expr)+ u"]] "+str(self.body) +u"]"
        return rval

    def evaluate(self,env):
        rval = None
        self.dbg_msg("eval-While stmt")
        while ( self.is_true_value ( self.expr.evaluate(env) ) 
                and not env.get_break_return() ):
            ## everytime of loop clear any continues
            env.clear_continue()
            self.dbg_msg("ifstmt: true condition")
            rval = self.body.evaluate( env )
        ## clear break if-any
        env.clear_break();
        self.dbg_msg("exiting While-stmt with rval="+str(rval))
        return rval

    def visit(self,walker):
        walker.visit_while_stmt(self)
        return

class DoWhileStmt(WhileStmt):
    """ do stmtlist  while ( exp )"""
    def __init__(self,expr,body,l,c,dbg=False):
        WhileStmt.__init__(self,expr,body,l,c,dbg)
        
    def __repr__(self):
        return u"[DoWhileStmt[expr="+str(self.expr)+u",body="+str(self.body)+"]]"
    
    def evaluate(self,env):
        """ first run is on the house, but then we keep count. Dog bites American style """
        rval = None
        first_time = True
        self.dbg_msg("eval-Do-While stmt")
        while ( first_time or self.is_true_value ( self.expr.evaluate(env) ) 
                and not env.get_break_return() ):
            ## everytime of loop clear any continues
            env.clear_continue()
            self.dbg_msg("ifstmt: true condition")
            rval = self.body.evaluate( env )
            first_time = False
        ## clear break if-any
        env.clear_break();
        self.dbg_msg("exiting Do-While-stmt with rval="+str(rval))
        return rval

class ForStmt(Stmt):
    """ For ( exp1 ; exp2 ; exp3 ) stmtlist  end"""
    def __init__(self,expr1,expr2, expr3,body,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.expr_init = expr1;
        self.expr_cond = expr2;
        self.expr_update = expr3;        
        self.body = body
        self.class_name = "ForStmt"
    
    def __repr__(self):
        rval = u"\t\n [ForStmt[[ ("+str(self.expr_init)+"; "+\
            str(self.expr_cond) + "; " +\
            str(self.expr_update)+") ]] " + str(self.body) +"]"
        return rval

    def evaluate(self,env):
        self.dbg_msg( "Eval-For-stmt: ")
        rval = None
        self.dbg_msg("eval-For-stmt")
        rval = self.expr_init.evaluate(env)        
        while ( self.is_true_value ( self.expr_cond.evaluate(env) )
                and not env.get_break_return() ):
            ## everytime of loop clear any continues
            env.clear_continue()
            rval = self.body.evaluate( env )
            # update happens after body evaluates - this is C-style            
            self.expr_update.evaluate( env )
        ## clear break if-any
        env.clear_break();
        self.dbg_msg(u"exiting For-stmt with rval="+str(rval))
        return rval

    def visit(self,walker):
        walker.visit_for_stmt(self)
        return

class AssignStmt(Stmt):
    """ lhs = rhs """
    def __init__(self,lvalue,op,rvalue,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.lvalue=lvalue
        self.assignop=op
        self.rvalue=rvalue
        self.class_name = "AssignStmt"

    def __repr__(self):
        return u"\n\t [AssignStmt[ "+ str(self.lvalue)+u"] " + \
               Token.token_types[self.assignop.kind] + \
               u"\t Expr [" + str(self.rvalue) + u"]]"

    def do_assignop(self, lvalue, rhs, kind, env):
        rval = None
        if ( kind == Token.EQUALS ):
            env.set_id( lvalue.id, rhs )
            rval = rhs
        else:
            raise Exception("Unknown assign operator @ "+self.get_pos())
        return rval
    
    def evaluate(self,env):
        if self.assignop.kind in Token.ASSIGNOP:
            self.dbg_msg(u"assignop: rhs = "+str(self.rvalue) )
            rhs = self.rvalue.evaluate(env)
            self.do_assignop(self.lvalue,
                             rhs,
                             self.assignop.kind, env)            
            self.dbg_msg(u"assignop lvalue ["+str(self.lvalue) \
                             +u"] = ["+str(rhs) + \
                             u"( saved as ) " + \
                             str(self.lvalue))
                             #str(env.get_id(self.lvalue.id)) )
            return rhs
        raise Exception("Unknown assign operator @ "+self.get_pos())
    
    def visit(self, walker):
        walker.visit_assign_stmt(self)
        return

class PrintStmt(Stmt):
    """ print EXPR """
    def __init__(self,exprlst, l, c, dbg):
        Stmt.__init__(self,l,c,dbg)
        self.exprlst = exprlst
        
    def __repr__(self):        
        return u"\n\t [PrintStmt[ "+ str(self.exprlst)+u"]]"

    def do_printop(self,env):
        val = self.exprlst.evaluate( env  ) 
        print(val) #this prints to output
        return val
    
    def evaluate(self,env):
        self.do_printop(env)
        return None
    
    def visit(self, walker):
        walker.visit_print_stmt(self)
        return

class EvalStmt(Stmt):
    """ EXPR """
    def __init__(self,expr, l, c, dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.expr = expr
        
    def __repr__(self):
        return u"\n\t [EvalStmt[ "+ str(self.expr)+u"/"+str((self.expr.__class__))+u"]]"

    def evaluate(self,env): 
        if ( self.debug ) : 
            print(u"evaluating EvalStmt : %s"%str(self.expr))
            print(self.expr.__class__)
        return self.expr.evaluate(env)

    def visit(self, walker):
        walker.visit_eval_stmt(self)
        return

## PLACEHOLDER
class ArgList:
    """ defines argument list in a function definition """
    def __init__(self,argvals, l, c, dbg=False):
        """ to get self.args, use get_list() method """
        self.args = argvals 
        self.line = l
        self.col = c
        self.parent = None
        
    def __len__(self):
        return len(self.args)

    def get_list(self):
        return self.args

    def __repr__(self):
        return "\n\t [ArgList["+ ",".join(map(str,self.args))+"]]"

    def visit(self,walker):
        walker.visit_arg_list(self)
        return

#TODO : derive from 'list' and 'Stmt' class and update code
class ValueList:
    """ defines value list in a function definition """
    def __init__(self,argvals, l, c, dbg =False):
        """ to get self.args, use get_list() method """
        self.args = argvals 
        self.debug = dbg
        self.line = l
        self.col = c
        self.parent = None
        
    def append(self,obj):
        self.args.append(obj)
        return
        
    def __len__(self):
        return len(self.args)
        
    def __getitem__(self,idx):
        """ index into the object like a list : @idx - caveat emptor """
        return self.args[idx]
    
    def get_list(self):
        return self.args

    #def evaluate(self):
    #    if ( isinstance(self.args,list) and len(self.args) == 1):
    #        return self.args[0]
    #    return self.args

    def __repr__(self):
        return "\n\t [ValueList["+ ",".join(map(str,self.args))+"]]"

    def visit(self,walker):
        walker.visit_value_list(self)
        return

    
class StmtList(Stmt):
    def __init__(self,stmt=[], dbg=False,istoplevel=False):
        Stmt.__init__(self,0,0,dbg)
        self.List = copy.copy(stmt)
        self.toplevel = istoplevel
        
    def __len__(self):
        return len(self.List)        
    
    def append(self,stmt_x):
        self.dbg_msg(u"==>"+str(stmt_x.__class__))
        self.dbg_msg(u"adding new statement " + str(stmt_x.__class__) )
        self.List.append(stmt_x)
        return
    
    def __repr__(self):
        rval = u"\t [StmtList[ "+ u"\n ".join(map(str,self.List)) + u"]]\n"
        return rval

    def evaluate(self,env):
        rval = None
        for stmt in self.List:
            self.dbg_msg(  u"STMTLIST => STMT" )
            if ( env.break_return_continue() ):
                break;
            self.dbg_msg( stmt.__class__ )
            rval = stmt.evaluate(env)
        return rval
    
    def visit(self,walker):
        """ visit stmt list method """
        if self.toplevel:
            walker.visit_program_or_script(self)
        else:
            walker.visit_stmt_list(self)
        return
    
class Function(Stmt):
    """ function definition itself """
    def __init__(self,fname,arglist,body,l,c,dbg=False):
        Stmt.__init__(self,l,c,dbg)
        self.name = fname
        self.arglist = arglist
        self.body = body
        self.dbg_msg( u"function "+fname+u" was defined" )
        
    def dbg_msg(self, msg):
        if ( self.debug ):
            print(u"## ",msg)
        return

    def __repr__(self):
        return u"\n\t [Function[ "+ str(self.name)+u"( " + \
            str(self.arglist) + u")]\n" + \
            u"\t Body [" + str(self.body) + u"]]\n"
    
    def evaluate(self,env):
        ## push stuff into the call-stack
        env.call_function(u"%s"%(self.name),u" at  %s"%(self.get_pos()))
        ## check arguments match, otherwise raise error
        args = env.get_args()#.get_list()
        fargs = self.arglist.get_list()
        if ( len(args) != len(fargs) ):
            raise Exception("Call Arguments donot match with" + \
                                "function definition @ "+self.get_pos())
        
        ## create local variables on the stack in order of definitions
        lut={}
        for idx in range(0,len(fargs)):
            varname = fargs[idx]
            value = args[idx]
            lut[varname]=value
        env.set_local(lut)
        
        ## invoke the function
        rval = self.body.evaluate(env)
        
        ## get the value from the stack.
        rval = env.get_retval( );

        env.clear_call()
        ## pop stuff into the call-stack
        env.return_function(self.name)
        return  rval

    def visit(self,walker):
        walker.visit_function(self)
        return



## -*- coding: utf-8 -*-
## (C) 2007, 2008, 2013, 2014, 2015 Muthiah Annamalai,
## Licensed under GPL Version 3
##
from __future__ import print_function

import argparse
from math import *
import copy
import  codecs
import os
import sys
import inspect
import string
import random ## builtins
from collections.abc import Callable
import tamil
from cmd import Cmd
from pprint import pprint
import time
import traceback
import json
import urllib
from functools import reduce
import importlib
import sys



PYTHON3 = (sys.version[0] == '3')
if PYTHON3:
    str = str
    string = str
    raw_input = input

tamil_sleep = time.sleep
tamil_date_time = time.asctime

######### E Z H I L - I N T E R P R E T E R ##########################
## scanner for exprs language
from .scanner import Token, Lexeme, Lex
from .tamil_scanner import tamilLex

## exceptions
from .errors import RuntimeException, ParseException

## runtime elements
from .runtime import  Environment, BuiltinFunction, \
 BlindBuiltins, DebugUtils, tamilCustomFunction

## AST elements
from .ast import Expr, ExprCall, ExprList, Stmt, ReturnStmt, \
 BreakStmt, ContinueStmt, ElseStmt, IfStmt, WhileStmt, \
 ForStmt, AssignStmt, PrintStmt, EvalStmt, ArgList, \
 ValueList, Function, StmtList, Identifier, Number, \
 String, Boolean

## Parser
from .ExprsParser import Parser
## Transforms
from .tamil_transforms import TransformEntryExitProfile, TransformSafeModeFunctionCheck,\
 TransformSemanticAnalyzer, TransformConstantFolder

from .tamil_serializer import SerializerXML
from .tamil_library import Load_URL_APIs

def tamil_keywords():
    keywords = {u"பதிப்பி":"PRINT",u"தேர்ந்தெடு":"SWITCH",
      u"தேர்வு":"CASE",u"ஏதேனில்":"OTHERWISE",u"ஆனால்":"IF",u"இல்லைஆனால்":"ELSEIF",
      u"இல்லை":"ELSE",u"ஆக":"FOR",u"ஒவ்வொன்றாக":"FOREACH",u"இல்":"COMMA",
      u"வரை":"WHILE",u"செய்":"DO",u"முடியேனில்":"DOWHILE",u"பின்கொடு":"RETURN",u"முடி":"END",
      u"நிரல்பாகம்":"DEF",u"தொடர்":"CONTINUE",u"நிறுத்து":"BREAK",u"உள்ளடக்கு":"IMPORT",
    "@":"ATRATEOF","=":"EQUALS","-":"MINUS",
    "+":"PLUS",">":"GT","<":"LT",">=":"GTEQ","<=":"LTEQ",
    "==":"EQUALITY","!=":"NEQ","*":"PROD","/":"DIV",",":"COMMA",
     "(":"LPAREN",")":"RPAREN","[":"LSQRBRACE","]":"RSQRBRACE",
     "{":"LCURLBRACE","}":"RCURLBRACE",
     ":":"COLON","%":"MOD","^":"EXP",
     "&&":"LOGICAL_AND",
     "&":"BITWISE_AND",
     "||":"LOGICAL_OR",
     "|":"BITWISE_OR",
     "<<":"BITWISE_LSHIFT",
     ">>":"BITWISE_RSHIFT",
     "~":"BITWISE_COMPLEMENT",
     "!":"LOGICAL_NOT"}
    val = keywords.keys()
    val.sort( tamil.utf8.compare_words_lexicographic )
    return val,[keywords[ref] for ref in val]

def tamil_quit():
    sys.exit(0)
    
def cmp(a, b):
    return (a > b) - (a < b)

def apply(func, args=(), kwargs={}):
    return func(*args, **kwargs)


def tamil_version():
    return 0.99

def tamil_copyright():
    return u"(C) 2007-2016 Muthiah Annamalai, and other contributors."

# you can also, get your name here, its easy!
def tamil_credits():
    return u"tamil language, version %g, was created by Muthiah Annamalai with several contributors."%(tamil_version())

def tamil_license():
    return u"Licensed under GPL Version 3"

def tamil_tamil_length(arg):
	return Number( len( tamil.utf8.get_letters( arg ) ) )

def tamil_getitem(*args):
    if len(args) < 2:
       raise Exception("getitem - too-few arguments, 2 or 3 arguments expected")
    with_default = (len(args) == 3)
    x = args[0]
    idx = args[1]
    if with_default:
        idxs = [idx,args[2]]
    else:
        idxs = [idx]
    #print("dispatching tamil getitem",type(x),type(idx),x,idx,x[idx])
    if ( isinstance(x,dict) ):
        return x.get(*idxs)
    elif ( isinstance(x,list) or hasattr( x, '__getitem__') ):
        return x.__getitem__(*idxs)
    return x.__getitem__(*idxs)

def tamil_islist( x ):
    return Boolean(isinstance(x,list))

def tamil_isnumber( x ):
    return Boolean(isinstance(x,Number) or type(x) in [int, float] )

def tamil_setitem(x,key,val):
    #print("dispatching tamil setitem",type(x),type(idx),x,idx,x[idx])
    #if ( isinstance(x,dict) or isinstance(x,list) or hasattr( x, '__setitem__') ):
    x.__setitem__(key,val)
    #else:
    #    x[key]=val
    return

def tamil_substr(str_var,start,end):
    return str_var[start:end]

def tamil_load(filename):
    # load the file into the interpreter
    # we have a symbol table and use it
    tamil_execute([filename])

class NoClobberDict(dict):
    """ dictionary structure with a set like mathematical structure.
        that way when you insert functions for tamil library, we don't trample on each other
        by accidentally overwriting stuff"""
    def __init__(self):
        dict.__init__(self)
    def __setitem__(self,key,val):
        if ( self.get(key,None) ):
            #print(u"-------> %s"%str(self.get(key)))
            traceback.print_stack()
            raise KeyError(u"Dictionary is getting clobbered; key '"+key+"' already present '%s'"%str(self.get(key)))
        dict.__setitem__(self,key,val)

global_interpreter = None

def tamil_list_functions(*args):
    global global_interpreter
    global_interpreter.list_functions(*args)
    
def tamil_eval(*args):
    global global_interpreter
    env = global_interpreter.env
    debug = global_interpreter.debug
    
    lexer = tamilLex(fname=[args[0]])
    if ( debug ): lexer.dump_tokens()
    parse_eval = Interpreter( lexer, debug, global_interpreter.SAFE_MODE,False,env=global_interpreter.env)
    try:
        web_ast = parse_eval.parse()
        parse_eval.call_stack.append("__tmptop__")
        return parse_eval.evaluate()
    except Exception as e:
        raise e
    return

def tamil_profile(*args):
    global global_interpreter
    env = global_interpreter.env
    if args[0] == u"begin":
        env.enable_profiling()
        return True
    
    if not env.profiler:
        raise RuntimeException(u"profile begin should be called before other profile commands like end, results etc")
    
    if args[0] == u"end":
        env.disable_profiling()
        return
    
    if args[0] == u"results":
        env.report_profiling()
        env.reset_profiling()
        return
    
    raise RuntimeException(u"profile command used with unknown argument %s"%args[0])

def tamil_execute(*args):
    global global_interpreter
    global_interpreter.reset()
    env = global_interpreter.env
    debug = global_interpreter.debug
    
    if len(args) < 1:
        raise RuntimeException(u"tamil_execute: missing filename argument")
    if (debug): print(args[0])
    lexer = tamilLex(fname=args[0])
    if ( debug ): lexer.dump_tokens()
    try:
        parse_eval = Interpreter( lexer, debug, global_interpreter.SAFE_MODE,False,env=global_interpreter.env)    
        web_ast = parse_eval.parse()
        parse_eval.call_stack.append("__evaluate__")
        return parse_eval.evaluate()
    except Exception as e:
        traceback.print_stack()
        raise e
    return

## Iyakki : tamil inge irunthu iyakkapadum
class Interpreter(DebugUtils):
    """ when you add new language feature, add a AST class 
    and its evaluate methods. Also add a parser method """
    def __init__(self,lexer, debug = False,safe_mode=False,update=True,stacksize=500,env=None):
        DebugUtils.__init__(self,debug)
        self.SAFE_MODE = safe_mode
        self.debug = debug
        self.MAX_REC_DEPTH = 10000 #for Python
        self.stacksize = stacksize #for tamil stacksize
        self.lexer=lexer
        self.ast=None
        self.function_map = NoClobberDict()#parsed functions
        self.builtin_map = NoClobberDict() #pointers to builtin functions
        self.call_stack = list() #call stack
        if not env:
            self.env = Environment( self.call_stack, self.function_map, \
                                        self.builtin_map, self.debug, self.stacksize)
        else:
            self.env = env
        sys.setrecursionlimit( self.MAX_REC_DEPTH ) # have a large enough Python stack
        ## use the default Exprs parser.
        lang_parser = Parser(self.lexer,self.function_map, \
                             self.builtin_map, self.debug )
        self.parser = lang_parser
        
        ## run the installation code last
        self.install_builtins()
        self.install_blind_builtins()
        
        if update :
            ## update globals
            global global_interpreter
            global_interpreter = self

    def get_fname(self):
        return self.lexer.fname
        
    def reset(self):
        """ reset lexer and parser """
        self.parser.reset()
        self.lexer.reset()
        return

    def change_parser(self,parser_ctor):
        """ change the parser to your-cool-language using
         a lambda function parser_ctor the constructor """
        lang_parser = parser_ctor(self.lexer,self.function_map, \
                             self.builtin_map, self.debug )
        self.parser = lang_parser
        return True
    
    def list_functions(self,*args):
        if ( not args ):
            user_prefix = None
        else:
            user_prefix = args[0]
                
        print(u"######### Builtin Function ########")
        if not user_prefix:
            pos = 0
            for xpos,fcn in enumerate(self.builtin_map.keys()):
                if ( user_prefix and not fcn.startswith(user_prefix) ):
                    continue
                pos = pos + 1
                print(u"%03d> %s"%(pos,str(fcn)))
            return
        
        print(u"######### Functions - Library + User defined - ########")
        pos = 0
        for xpos,fcn in enumerate(self.function_map.keys()):
            if ( user_prefix and not fcn.startswith(user_prefix) ):
                continue
            pos = pos + 1
            print(u"%03d> %s"%(pos,str(fcn)))
        
    def add_blind_fcns(self, bfn, b):
        """ an internal method to reduce repetition """
        if ( not inspect.ismethod(bfn) 
             and not inspect.isclass(bfn)
             and isinstance(bfn, Callable) ):
            self.dbg_msg("adding: "+b);
            self.builtin_map[b] = BlindBuiltins( bfn, b);
        else:
            self.dbg_msg("skipping: "+b);
        return

    def install_blind_builtins(self):
        """ we dont know the arity of these functions,
        hence we call-em the blind builtins, but we 
        still add them to the builtin_map """
        
        for b in dir(os):
            if self.SAFE_MODE:
                break
            bfn = getattr( os ,b)
            if str(b).find('exec') >= 0 or str(b).find('spawn') >= 0:
                continue
            if PYTHON3 and b == u"replace":
                continue
            self.add_blind_fcns( bfn, b)
        
        for b in dir(sys):
            bfn = getattr( sys ,b)
            self.add_blind_fcns( bfn, b)
        
    @staticmethod
    def tamil_assert( x  ):
        # we raise Exception(u'Assertion failed!') further up stack
        if not x:
            # TBD: print the description of the AST x
            pass
        assert x
        return True
    
    # file IO functions - 6 total
    @staticmethod
    def file_open(*args):
        try:
            args = list(args)
            if ( len(args) < 2 ):
                args.append('r') #default op = read
            if ( len(args) < 3 ):
                args.append('utf-8') #default encoding = utf-8
            fp = codecs.open(*args)
            return fp
        except Exception as e:
            print(u"WARNING : cannot open file %s with exception %s",args[0],str(e))
        return -1
    
    @staticmethod
    def file_close(*args):
        try:
            assert( len(args) == 1 )
            assert( hasattr(args[0],'close') )
            args[0].close()
        except Exception as e:
            print(u"WARNING: file_close failed with exception %s"%str(e))
            return -1
        return 0
    
    @staticmethod
    def file_read(*args):
        assert( len(args) == 1 )
        assert( hasattr(args[0],'read') )
        return String( args[0].read() )
    
    @staticmethod
    def file_readlines(*args):
        assert( len(args) == 1 )
        assert( hasattr(args[0],'readlines') )
        return args[0].readlines()
    
    @staticmethod
    def file_write(*args):
        assert( len(args) == 2 )
        assert( hasattr(args[0],'write') )
        return args[0].write( args[1] )
    
    @staticmethod
    def file_writelines(*args):
        assert( len(args) == 2 )
        assert( hasattr(args[0],'writelines') )
        return args[0].writelines(args[1])
    
    # marshalling            
    @staticmethod
    def RAWINPUT(args):
        try:
            op = tamilCustomFunction.raw_input(args)
        except KeyboardInterrupt as e :
            print(u"\nTo exit the program type : exit or press CTRL + D on your keyboard")
            return String("")
        return String(op)
    
    @staticmethod
    def INPUT(args):
        try:
            op = eval(tamilCustomFunction.raw_input(args))
        except KeyboardInterrupt as e :
            print(u"\nTo exit the program type : exit or press CTRL + D on your keyboard")
            return Number(0)
        if ( isinstance(op,int) or isinstance(op,float) ):
            return Number(0.0+op)
        return String( op )
    
    @staticmethod   
    def SPRINTF_worker(*args):
        if ( len(args) < 1 ):
            raise Exception(u'Not enough arguments to printf() function')
        fmt = args[0]
        arg = tuple( args[1:] );
        opstr = fmt%arg;
        return opstr
    
    @staticmethod
    def SPRINTF(*args):
        opstr = Interpreter.SPRINTF_worker(*args);
        return String(opstr)
    
    @staticmethod   
    def PRINTF(*args):
        str_op = Interpreter.SPRINTF_worker(*args);
        print(str_op)
        return Number(len(str_op))
    
    @staticmethod
    def tamil_reverse(*args):
       if ( len(args) != 1 ): raise Exception('One argument alone expected to reverse function')
       if ( isinstance( args[0], str ) or isinstance( args[0], str) ):
           return String(args[0][::-1]) #string-reverse
       
       return list.reverse(args[0])
    
    @staticmethod
    def tamil_pause(*args):
        if ( len(args) >= 1 ):
             print(args[0])
        if ( len(args) < 2 ):
             tamil_sleep( 5 )
        else:
             tamil_sleep( args[1] )
        return
    
    def add_builtin(self,call_name,call_handle,nargin=1,ta_alias=None):        
        # make sure you don't clobber yourself
        assert not self.builtin_map.get(call_name,None)
        if ( nargin == -1 ):
            self.builtin_map[call_name] = BlindBuiltins( call_handle, call_name, self.debug)
        else:
            self.builtin_map[call_name] = BuiltinFunction( call_handle, call_name, nargin, self.debug )
        # update the alias if something was supplied
        if ( ta_alias ):
            assert not self.builtin_map.get(ta_alias,None) #noclobber
            self.builtin_map[ta_alias] = self.builtin_map[call_name]
        
        return True
    
    def install_builtins(self):
        """ populate with the builtin functions"""
        self.builtin_map[u'printf']=BlindBuiltins(Interpreter.PRINTF,u'printf',self.debug)
        self.builtin_map[u'sprintf']=BlindBuiltins(Interpreter.SPRINTF,u'sprintf',self.debug)
        
        self.builtin_map['abs']=BlindBuiltins(abs,'abs',self.debug)
        self.builtin_map['all']=BlindBuiltins(all,'all',self.debug)
        self.builtin_map['any']=BlindBuiltins(any,'any',self.debug)
        
        if not PYTHON3:
            
            self.builtin_map['apply'] = BlindBuiltins(apply, 'apply', self.debug)
            self.builtin_map['memoryview']=BlindBuiltins(memoryview,'memoryview',self.debug)
            self.builtin_map['cmp']=BlindBuiltins(cmp,'cmp',self.debug)
        self.builtin_map['basestring']=BlindBuiltins(str,'basestring',self.debug)
        self.builtin_map['bin']=BlindBuiltins(bin,'bin',self.debug)
        self.builtin_map['bool']=BlindBuiltins(bool,'bool',self.debug)
        self.builtin_map['bytearray']=BlindBuiltins(bytearray,'bytearray',self.debug)
        self.builtin_map['bytes']=BlindBuiltins(bytes,'bytes',self.debug)
        self.builtin_map['callable']=BlindBuiltins(callable,'callable',self.debug)
        self.builtin_map['chr']=BlindBuiltins(chr,'chr',self.debug)
        #self.builtin_map['classmethod']=BlindBuiltins(classmethod,'classmethod',self.debug)
        #self.builtin_map['coerce']=BlindBuiltins(coerce,'coerce',self.debug)
        self.builtin_map['compile']=BlindBuiltins(compile,'compile',self.debug)
        self.builtin_map['complex']=BlindBuiltins(complex,'complex',self.debug)
        self.builtin_map['xor'] = BlindBuiltins(lambda *x: reduce(lambda a,b: a^b,x),'xor',self.debug)
        # special tamil commands
        self.builtin_map['copyright']=BlindBuiltins(tamil_copyright,'copyright',self.debug)
        self.builtin_map['version']=BlindBuiltins(tamil_version,'version',self.debug) 
        self.builtin_map['credits']=BlindBuiltins(tamil_credits,'credits',self.debug)
        
        # global interpreter
        self.builtin_map['dir']=BlindBuiltins(tamil_list_functions,'dir',self.debug)

        self.builtin_map['delattr']=BlindBuiltins(delattr,'delattr',self.debug)
        self.builtin_map['dict']=BlindBuiltins(dict,'dict',self.debug)
        self.builtin_map['python_dir']=BlindBuiltins(dir,'python_dir',self.debug)
        self.builtin_map['divmod']=BlindBuiltins(divmod,'divmod',self.debug)
        self.builtin_map['enumerate']=BlindBuiltins(enumerate,'enumerate',self.debug)
        
        # skip these system functions
        self.builtin_map['eval']=BlindBuiltins(tamil_eval,'eval',self.debug)
        self.builtin_map[u'மதிப்பீடு']=self.builtin_map['eval']
        self.builtin_map['execute']=BlindBuiltins(tamil_execute,'execute',self.debug)
        self.builtin_map[u'இயக்கு']=self.builtin_map['execute']
        
        self.builtin_map['profile'] = BlindBuiltins(tamil_profile,'profile',self.debug)
        #self.builtin_map['execfile']=BlindBuiltins(execfile,'execfile',self.debug)
        #self.builtin_map['exit']=BlindBuiltins(exit,'exit',self.debug)
        if not PYTHON3:
            self.builtin_map['file']=BlindBuiltins(open,'file',self.debug)
            self.builtin_map['intern'] = BlindBuiltins(sys.intern, 'intern', self.debug)

        
        try:
            self.builtin_map['filter']=BlindBuiltins(filter,'filter',self.debug)
            self.builtin_map['float']=BlindBuiltins(float,'float',self.debug)
            self.builtin_map['format']=BlindBuiltins(format,'format',self.debug)
            self.builtin_map['frozenset']=BlindBuiltins(frozenset,'frozenset',self.debug)
            self.builtin_map['getattr']=BlindBuiltins(getattr,'getattr',self.debug)
            #self.builtin_map['globals']=BlindBuiltins(globals,'globals',self.debug)
            self.builtin_map['hasattr']=BlindBuiltins(hasattr,'hasattr',self.debug)
            self.builtin_map['hash']=BlindBuiltins(hash,'hash',self.debug)
            #self.builtin_map['help']=BlindBuiltins(help,'help',self.debug)
            self.builtin_map['hex']=BlindBuiltins(hex,'hex',self.debug)
            self.builtin_map['id']=BlindBuiltins(id,'id',self.debug)        
            self.builtin_map['int']=BlindBuiltins(int,'int',self.debug)
            self.builtin_map['isinstance']=BlindBuiltins(isinstance,'isinstance',self.debug)
            self.builtin_map['issubclass']=BlindBuiltins(issubclass,'issubclass',self.debug)
            self.builtin_map['iter']=BlindBuiltins(iter,'iter',self.debug)
            self.builtin_map['len']=BlindBuiltins(len,'len',self.debug)
            self.builtin_map['license']=BlindBuiltins(tamil_license,'license',self.debug)
            self.builtin_map['long']=BlindBuiltins(int,'long',self.debug)
            self.builtin_map['map']=BlindBuiltins(map,'map',self.debug)
            #self.builtin_map['max']=BlindBuiltins(max,'max',self.debug)
        except Exception as e:
            pass
        
        try:
            self.builtin_map['memoryview']=BlindBuiltins(memoryview,'memoryview',self.debug)
        except NameError as ie:
            if(self.debug): 
                print(u"Name Error:",str(ie))
    
        #self.builtin_map['min']=BlindBuiltins(min,'min',self.debug)
        self.builtin_map['next']=BlindBuiltins(next,'next',self.debug)
        self.builtin_map['object']=BlindBuiltins(object,'object',self.debug)
        self.builtin_map['oct']=BlindBuiltins(oct,'oct',self.debug)
        #self.builtin_map['open']=BlindBuiltins(open,'open',self.debug)
        self.builtin_map['ord']=BlindBuiltins(ord,'ord',self.debug)
        self.builtin_map['substr']=BlindBuiltins(tamil_substr,'substr',self.debug)
        #self.builtin_map['pow']=BlindBuiltins(pow,'pow',self.debug)
        self.builtin_map['property']=BlindBuiltins(property,'property',self.debug)
        self.builtin_map['quit']=BlindBuiltins(tamil_quit,'quit',self.debug)
        self.builtin_map['range']=BlindBuiltins(range,'range',self.debug)        
        if not PYTHON3:
            self.builtin_map['reduce']=BlindBuiltins(reduce,'reduce',self.debug)
            self.builtin_map['reload']=BlindBuiltins(importlib.reload,'reload',self.debug)
            self.builtin_map['str']=BlindBuiltins(str,'str',self.debug)
        
        self.builtin_map['repr']=BlindBuiltins(repr,'repr',self.debug)
        self.builtin_map['reversed']=BlindBuiltins(reversed,'reversed',self.debug)
        self.builtin_map['round']=BlindBuiltins(round,'round',self.debug)
        self.builtin_map['set']=BlindBuiltins(set,'set',self.debug)
        self.builtin_map['setattr']=BlindBuiltins(setattr,'setattr',self.debug)
        self.builtin_map['slice']=BlindBuiltins(slice,'slice',self.debug)
        self.builtin_map['sorted']=BlindBuiltins(sorted,'sorted',self.debug)
        self.builtin_map['staticmethod']=BlindBuiltins(staticmethod,'staticmethod',self.debug)
        self.builtin_map['sum']=BlindBuiltins(sum,'sum',self.debug)
        self.builtin_map['super']=BlindBuiltins(super,'super',self.debug)
        self.builtin_map['tuple']=BlindBuiltins(tuple,'tuple',self.debug)
        self.builtin_map['type']=BlindBuiltins(type,'type',self.debug)
        self.builtin_map['unichr']=BlindBuiltins(chr,'unichr',self.debug)
        self.builtin_map['vars']=BlindBuiltins(vars,'vars',self.debug)
        if not PYTHON3:
            self.builtin_map['range']=BlindBuiltins(range,'range',self.debug)
        self.builtin_map['zip']=BlindBuiltins(zip,'zip',self.debug)
        
        # common/generic functions
        self.builtin_map['__getitem__']=BlindBuiltins(tamil_getitem,"__getitem__")
        self.builtin_map['__setitem__']=BuiltinFunction(tamil_setitem,"__setitem__",3)
        
        #file-IO functions
        self.builtin_map["file_open"]=BlindBuiltins(Interpreter.file_open,"file_open")
        self.builtin_map["file_close"]=BuiltinFunction(Interpreter.file_close,"file_close")
        self.builtin_map["file_read"]=BuiltinFunction(Interpreter.file_read,"file_read")
        self.builtin_map["file_readlines"]=BuiltinFunction(Interpreter.file_readlines,"file_readlines")
        self.builtin_map["file_write"]=BuiltinFunction(Interpreter.file_write,"file_write",2)
        self.builtin_map["file_writelines"]=BuiltinFunction(Interpreter.file_writelines,"file_writelines",2)
        
        # input statements
        self.builtin_map["input"]=BuiltinFunction(Interpreter.INPUT,"input")
        self.builtin_map["raw_input"]=BuiltinFunction(Interpreter.RAWINPUT,"raw_input")
        
        # keywords
        self.add_builtin("keywords",lambda : tamil_keywords()[0],nargin=0,ta_alias=u"குறிசொற்கள்")
        
        # assert
        self.builtin_map["assert"]=BuiltinFunction(Interpreter.tamil_assert,"assert")
        self.builtin_map[u"உறுதிப்படுத்த"]=self.builtin_map["assert"]
        
        # str - used for simple serialization in tamil
        self.builtin_map["str"]=BuiltinFunction(str,"str")
        
        # sys_platform - used for Windows/Linux identification
        self.builtin_map["sys_platform"]=BuiltinFunction(lambda : sys.platform,"sys_platform",padic=0)
        
        # sleep/pause
        self.add_builtin("sleep",tamil_sleep,nargin=1,ta_alias=u"உரங்கு")
        self.builtin_map["pause"]=BlindBuiltins(Interpreter.tamil_pause,"pause")

        # date/time
        self.add_builtin("date_time",tamil_date_time,nargin=0,ta_alias=u"தேதி_நேரம்")
        self.add_builtin("time",time.time,nargin=0,ta_alias=u"நேரம்")
        self.add_builtin("ctime",time.ctime,nargin=1,ta_alias=u"cநேரம்")
        self.add_builtin("clock",time.time,nargin=0)

        # islist, isnumber predicates
        self.add_builtin("islist",tamil_islist,nargin=1,ta_alias=u"பட்டியலா")
        self.add_builtin("isnumber",tamil_isnumber,nargin=1,ta_alias=u"எண்ணா")
        # random functions
        aslist = True;
        self.builtin_map["choice"]=BlindBuiltins(random.choice,"choice",self.debug,aslist)
        self.builtin_map["random"]=BuiltinFunction(random.random,"random",0)
        self.builtin_map["seed"]=BuiltinFunction(random.seed,"seed")
        self.builtin_map["randint"]=BuiltinFunction(random.randint,"randint",2)
    
        # math functions
        self.builtin_map["acos"]=BuiltinFunction(acos,"acos")
        self.builtin_map["asin"]=BuiltinFunction(asin,"asin")
        self.builtin_map["atan"]=BuiltinFunction(atan,"atan")
        self.builtin_map["atan2"]=BuiltinFunction(atan2,"atan2")
        self.builtin_map["ceil"]=BuiltinFunction(ceil,"ceil")
        self.builtin_map["cos"]=BuiltinFunction(cos,"cos")
        self.builtin_map["cosh"]=BuiltinFunction(cosh,"cosh")
        self.builtin_map["degrees"]=BuiltinFunction(degrees,"degrees")
        self.builtin_map["e"]=BuiltinFunction(lambda : e,"e",0)
        self.builtin_map["exp"]=BuiltinFunction(exp,"exp")
        self.builtin_map["fabs"]=BuiltinFunction(fabs,"fabs")
        self.builtin_map["floor"]=BuiltinFunction(floor,"floor")
        self.builtin_map["fmod"]=BuiltinFunction(fmod,"fmod",2)
        self.builtin_map["frexp"]=BuiltinFunction(frexp,"frexp")
        self.builtin_map["hypot"]=BuiltinFunction(hypot,"hypot",2)
        self.builtin_map["ldexp"]=BuiltinFunction(ldexp,"ldexp")
        self.builtin_map["log"]=BuiltinFunction(log,"log")
        self.builtin_map["log10"]=BuiltinFunction(log10,"log10")
        self.builtin_map["modf"]=BuiltinFunction(modf,"modf",2)
        self.builtin_map["pi"]=BuiltinFunction(lambda : pi,"pi",0)
        self.builtin_map["pow"]=BuiltinFunction(pow,"pow",2)
        self.builtin_map["radians"]=BuiltinFunction(radians,"radians")
        self.builtin_map["sin"]=BuiltinFunction(sin,"sin")
        self.builtin_map["sinh"]=BuiltinFunction(sinh,"sinh")
        self.builtin_map["sqrt"]=BuiltinFunction(sqrt,"sqrt")
        self.builtin_map["tan"]=BuiltinFunction(tan,"tan")
        self.builtin_map["tanh"]=BuiltinFunction(tanh,"tanh")

        self.builtin_map["max"]=BuiltinFunction(max,"max",2)
        self.builtin_map["min"]=BuiltinFunction(min,"min",2)
        # turtle functions - optional - 
        try:
            from TamilTurtle import TamilTurtle

            turtle_attrib = TamilTurtle.functionAttributes();
            for nargs,fcnName in list(turtle_attrib.items()):
                for vv in fcnName:
                    turtlefcn = u"turtle_"+vv;
                    if ( self.debug ): print(nargs, vv)
                    if ( nargs == -1 ):
                        self.builtin_map[turtlefcn] = BlindBuiltins(getattr(TamilTurtle, vv),vv,self.debug)
                    else:
                        self.builtin_map[turtlefcn] = BuiltinFunction( getattr( TamilTurtle, vv ), turtlefcn, nargs )
        except ImportError as ie:
            if ( self.debug ): 
                print(u"Cannot Import TamilTurtle module; ignoring for now")
        # all builtin functions should pass the following test:
        # for attr in dir(module):
        #     assert callable( getattr( module, attr ) )
        # non-callable attributes cannot be in the builtins list        
        # string module builtins
        self.builtin_map["find"] = BuiltinFunction(string.find,"find",2)
        if not PYTHON3:
            self.builtin_map["atof"] = BuiltinFunction(string.atof,"atof",1)
            self.builtin_map["atof_error"] = BuiltinFunction(string.atof_error,"atof_error",1)
            self.builtin_map["atoi"] = BuiltinFunction(string.atoi,"atoi",1)
            self.builtin_map["atoi_error"] = BuiltinFunction(string.atoi_error,"atoi_error",1)
            self.builtin_map["atol"] = BuiltinFunction(string.atol,"atol",1)
            self.builtin_map["atol_error"] = BuiltinFunction(string.atol_error,"atol_error",1)
            self.builtin_map["capitalize"] = BuiltinFunction(string.capitalize,"capitalize",1)
            self.builtin_map["capwords"] = BuiltinFunction(string.capwords,"capwords",1)
            self.builtin_map["center"] = BuiltinFunction(string.center,"center",1)
            self.builtin_map["count_string"] = BuiltinFunction(string.count,"count",1)
            self.builtin_map["expandtabs"] = BuiltinFunction(string.expandtabs,"expandtabs",1)

        self.builtin_map["index_string"] = BuiltinFunction(string.index,"index",2)
        if not PYTHON3:
            self.builtin_map["index_error"] = BuiltinFunction(string.index_error,"index_error",1)
        self.builtin_map["join"] = BuiltinFunction(str.join,"join",1)
        if not PYTHON3:
            self.builtin_map["joinfields"] = BuiltinFunction(string.joinfields,"joinfields",1)
        self.builtin_map["ljust"] = BuiltinFunction(string.ljust,"ljust",1)
        self.builtin_map["lower"] = BuiltinFunction(string.lower,"lower",1)
        self.builtin_map["lstrip"] = BuiltinFunction(string.lstrip,"lstrip",1)
        if not PYTHON3:
            self.builtin_map["maketrans"] = BuiltinFunction(string.maketrans,"maketrans",1)
        self.builtin_map["replace"] = BuiltinFunction(string.replace,"replace",3)
        self.builtin_map["rfind"] = BuiltinFunction(string.rfind,"rfind",2)
        self.builtin_map["rindex"] = BuiltinFunction(string.rindex,"rindex",1)
        self.builtin_map["rjust"] = BuiltinFunction(string.rjust,"rjust",1)
        self.builtin_map["rsplit"] = BuiltinFunction(string.rsplit,"rsplit",1)
        self.builtin_map["rstrip"] = BuiltinFunction(string.rstrip,"rstrip",1)
        self.builtin_map["split"] = BuiltinFunction(string.split,"split",2)
        if not PYTHON3:
            self.builtin_map["splitfields"] = BuiltinFunction(string.splitfields,"splitfields",1)
        self.builtin_map["strip"] = BuiltinFunction(string.strip,"strip",1)
        self.builtin_map["swapcase"] = BuiltinFunction(str.swapcase,"swapcase",1)
        self.builtin_map["translate"] = BuiltinFunction(str.translate,"translate",1)
        self.builtin_map["upper"] = BuiltinFunction(str.upper,"upper",1)
        self.builtin_map["zfill"] = BuiltinFunction(str.zfill,"zfill",2)
        # get/set methods are handled by generic __getitem__ and __setitem__
        
        # list methods - first argument, when required, is always a list obj
        self.builtin_map["append"] = BuiltinFunction(list.append,"append",2)
        self.builtin_map["insert"] = BuiltinFunction(list.insert,"insert",3)
        self.builtin_map["index"] = BuiltinFunction(list.index,"index",2)
        self.builtin_map["list"] = BuiltinFunction(list,"list",0)
        self.builtin_map["pop"] = BuiltinFunction(list.pop,"pop",1)
        
        #if PYTHON3:
        #    self.builtin_map["sort"] = BuiltinFunction(lambda x: sorted(x),"sort",1)
        #else:
        self.builtin_map["sort"] = BuiltinFunction(list.sort,"sort",1)
        self.builtin_map["count"]= BuiltinFunction(list.count,"count",2)
        self.builtin_map["extend"]= BuiltinFunction(list.extend,"extend",2)
        self.builtin_map["reverse"]= BuiltinFunction(Interpreter.tamil_reverse,"reverse",1)
        self.builtin_map["pop_list"] = BuiltinFunction(list.pop,"pop",1)
        self.builtin_map[u"மேல்அழி"] = self.builtin_map["pop_list"]
        # dictionary methods - first argument, when required, is always a dict obj
        self.builtin_map["clear"]= BuiltinFunction(dict.clear,"clear",1)
        self.builtin_map["copy"]= BuiltinFunction(dict.copy,"copy",1)
        self.builtin_map["fromkeys"]= BuiltinFunction(dict.fromkeys,"fromkeys",1)
        self.builtin_map["items"]= BuiltinFunction(dict.items,"items",1)
        
        if not PYTHON3:
            self.builtin_map["has_key"]= BuiltinFunction(dict.has_key,"has_key",1)
            self.builtin_map["iteritems"]= BuiltinFunction(dict.iteritems,"iteritems",1)
            self.builtin_map["iterkeys"]= BuiltinFunction(dict.iterkeys,"iterkeys",1)
            self.builtin_map["itervalues"]= BuiltinFunction(dict.itervalues,"itervalues",1)
            self.builtin_map["pop_dict"]= BuiltinFunction(dict.pop,"pop",1)
            self.builtin_map["popitem"]= BuiltinFunction(dict.popitem,"popitem",1)
            self.builtin_map["setdefault"]= BuiltinFunction(dict.setdefault,"setdefault",1)
        
        if PYTHON3:
            self.builtin_map["keys"]= BuiltinFunction(lambda x: list(dict.keys(x)),"keys",1)
            self.builtin_map["values"]= BuiltinFunction(lambda x: list(dict.values(x)),"values",1)
        else:
            self.builtin_map["keys"]= BuiltinFunction(dict.keys,"keys",1)
            self.builtin_map["values"]= BuiltinFunction(dict.values,"values",1)
        
        self.builtin_map["update"]= BuiltinFunction(dict.update,"update",1)
        
        # URL API's - load except in safe mode
        if not self.SAFE_MODE:
            Load_URL_APIs( self )
        
        # open-tamil API
        # get tamil letters
        self.add_builtin("get_tamil_letters",tamil.utf8.get_letters,nargin=1,ta_alias=u"தமிழ்_எழுத்துக்கள்")
        self.add_builtin(u"த",tamil.utf8.get_letters,nargin=1)
        self.add_builtin("tamil_length",tamil_tamil_length,nargin=1,ta_alias=u"தநீளம்")     
        
        # functions returning constant list of Tamil strings
        self.add_builtin("tamil_letters",lambda :  tamil.utf8.tamil_letters,
                         nargin=0,ta_alias=u"தமிழ்எழுத்து")
        self.add_builtin("tamil_uyir",lambda :  tamil.utf8.uyir_letters,
                         nargin=0,ta_alias=u"உயிர்எழுத்து")
        self.add_builtin("tamil_mei",lambda: tamil.utf8.mei_letters,
                         nargin=0,ta_alias=u"மெய்எழுத்து") 
        self.add_builtin("tamil_kuril",lambda: tamil.utf8.kuril_letters,
                         nargin=0,ta_alias=u"குரில்எழுத்து") 
        self.add_builtin("tamil_nedil",lambda: tamil.utf8.nedil_letters,
                         nargin=0,ta_alias=u"நேடில்எழுத்து") 
        self.add_builtin("tamil_vallinam",lambda: tamil.utf8.vallinam_letters,
                         nargin=0,ta_alias=u"வல்லினம்எழுத்து") 
        self.add_builtin("tamil_mellinam",lambda: tamil.utf8.mellinam_letters,
                         nargin=0,ta_alias=u"மெல்லினம்") 
        self.add_builtin("tamil_idayinam",lambda: tamil.utf8.idayinam_letters,
                         nargin=0,ta_alias=u"இடைனம்எழுத்து") 
        self.add_builtin("tamil_agaram",lambda: tamil.utf8.agaram_letters,
                         nargin=0,ta_alias=u"அகரம்எழுத்து") 
        self.add_builtin("tamil_uyirmei",lambda: tamil.utf8.uyirmei_letters,
                         nargin=0,ta_alias=u"உயிர்மெய்எழுத்து")

        self.add_builtin("tamil_istamil_prefix",tamil.utf8.istamil_prefix,
                         nargin=1)
        self.add_builtin("tamil_all_tamil", tamil.utf8.all_tamil,
                         nargin=1,ta_alias=u"தனித்தமிழா")
        self.add_builtin("tamil_hastamil",tamil.utf8.has_tamil,
                         nargin=1,ta_alias=u"தமிழ்கொண்டதா")
        self.add_builtin("tamil_reverse_word",tamil.utf8.reverse_word,
                         nargin=1,ta_alias=u"அந்தாதிமாற்று")
        return True

    def __del__(self):
        if self.debug:
            print("deleting Interpreter...")
        
    def __repr__(self):
        rval =  u"[Interpreter: "
        rval = rval + u"[Functions["
        for k in list(self.function_map.keys()):
            rval = rval + u"\n "+ str(self.function_map[k]) 
        rval = rval +u"]] "+ str(self.ast) +u"]\n"
        return rval

    def parse(self):
        """
        parser routine """
        self.ast = self.parser.parse()
        return self.ast
    
    def evaluate(self):
        try:
            if self.SAFE_MODE:
                TransformSafeModeFunctionCheck(interpreter=self,debug=self.debug)

            if self.SAFE_MODE:
                self.env.record_start_time()

            # always verify semantics before running code
            #if ( self.debug ): print(self.ast )
            if ( self.debug ): print("pre-semantic xform analyzer")
            TransformSemanticAnalyzer(interpreter=self, debug=self.debug)
            if ( self.debug ): print("post-semantic xform analyzer")
            # constant fold
            #for i in range(0,2):
            #    TransformConstantFolder( interpreter = self, debug=self.debug)
            
            if ( len(self.call_stack) == 0 ):
                self.env.call_function("__toplevel__") ##some context
            return self.ast.evaluate(self.env)
        except Exception as e:
            self.env.unroll_stack()
            if ( self.debug ):
                traceback.print_tb(sys.exc_info()[2])
            raise e
        
    def evaluate_interactive(self,env=None,throw=True):
        ## use a persistent environment for interactive interpreter
        try:
            ## use this from the interactive-interpreter
            rval = self.ast.evaluate(self.env)
            return [ rval, self.env ]
        #except RuntimeException as e:
        #    self.env.unroll_stack()
        #    #self.env.call_stack.pop()
        except Exception as e:
            self.env.unroll_stack()
            #self.env.call_stack.pop()
            print("ERROR : %s"%str(e))
            if throw:
                raise RuntimeException(e)
            if (self.debug): print(str(self.env))
        return [str(e),self.env]
        
# world-famous REPL
# derive from the Python standard library - Cmd
# refactor out REPL for tamil and exprs
class REPL(Cmd):
    def __init__(self,lang, lexer, parse_eval, debug=False):
        """ @lexer is language lexical analyzer
            @parse_eval is the language interpreter with builtins, runtime, parser etc..
            @lang is a descriptive string,
            @debug the boolean """
        Cmd.__init__(self)
        ## ala-Python like
        self.banner = u"""எழில் - ஒரு தமிழ் நிரலாக்க மொழி (Thu Aug 20 00:08:00 PDT 2017)
tamil : A Tamil Programming Language - version %g, (C) 2007-2017
Type "help", "copyright", "credits" or "license" for more information."""%tamil_version()
        
        self.lang = lang
        self.lexer = lexer
        self.parse_eval = parse_eval
        self.debug = debug
        self.line_no = 1
        self.env = None ## get the first instance from evaluate_interactive
        self.prevlines = u'' ## support for continued lines
        self.cmdloop()
        
    def parseline(self,line):
        arg,cmd = u"",u""
        if not PYTHON3:
            line = line.decode("utf-8")
        line=line.strip()
        if line in [u"exit",u"help",u"EOF",u"copyright",u"credits",u"license"]:
            cmd = line
            line = line+u"()"
        return [cmd,arg,line]
    
    def update_prompt(self):
        self.prompt = u"%s %d>> "%(self.lang,self.line_no)

    def continuedline(self):
        return len(self.prevlines) > 0
        
    def preloop(self):
        if not self.continuedline():
            self.update_prompt()
        print(u"%s"%(self.banner))
    
    def emptyline(self):
        pass
    
    def default(self,line):
        """ REPL is carried out primarily through this callback from the looping construct """
        self.line_no += 1
        #line = str(line)
        if ( self.debug ): print("evaluating line", line)
        if ( line == u'exit()' ): self.exit_hook(doExit=True)
        try:
            self.lexer.set_line_col([self.line_no, 0])
            if len(self.prevlines) > 0:
                total_line = self.prevlines + u'\n' + line
            else:
                total_line = line
            self.lexer.tokenize(total_line)
            #self.lexer.warn_incomplete_tokens()
            nbraces = list()
            COMPLEMENT = {'}':'{',')':'(',']':'['}
            for TKN_idx in range(len(self.lexer.tokens)-1,-1,-1):
                TKN = self.lexer.tokens[ TKN_idx ]
                if ( self.debug ): print(TKN, TKN_idx)
                if TKN.val in ["(", "[", "{"]:
                    nbraces.append(TKN.val)
                elif TKN.val in [")","]","}"]:
                    if len(nbraces) > 0 and nbraces[-1] == COMPLEMENT[TKN.val]:
                        del nbraces[-1]
                        if ( self.debug ): print("matched braces..")
                else:
                    if ( self.debug ): print("nota brace token",TKN.val)
            if len(nbraces) > 0:
                if ( self.debug ): 
                    print ("Mismatched braces")
                    for idx in nbraces:
                        print(idx)
                self.prevlines += u'\n' + line
                self.lexer.reset() #reset lexer
                return
            else:
                if ( self.debug ): print ("Braces matched")
                self.prevlines = u''
            [line_no,c] = self.lexer.get_line_col( 0 )
            if ( self.debug ): self.lexer.dump_tokens()
            
            self.prevlines = u'' #erase prev
            self.parse_eval.parse()
            if ( self.debug ):
                print(u"*"*60);  
            TransformSemanticAnalyzer( interpreter = self.parse_eval, debug=self.debug)
            #TransformConstantFolder( interpreter = self.parse_eval, debug=self.debug)
            
            [rval, self.env] = self.parse_eval.evaluate_interactive(self.env)
            if self.debug:
                print( u"return value", str(rval) )
            if hasattr( rval, 'evaluate' ):
                pprint(rval.__str__())
            elif hasattr(rval,'__str__'): #print everything except a None object
                pprint( rval )
        except Exception as excep:
            print(u"Exception in code, at line %d,  \"%s\" \n >>>>>>> %s "%(self.line_no-1,line,str(excep)))
            self.lexer.reset()
            if ( self.debug ):
                raise excep
                ## clear tokens in lexer
        if ( self.debug ): print(self.env)
        self.lexer.tokens = list()
        self.update_prompt()
    
    def do_EOF(self,line):
        print(u"\n")
        self.exit_hook()
        return True
    
    def exit_hook(self,doExit=False):
        if ( self.lang == u"எழில்"):
            print(u"******* வணக்கம்! பின்னர் உங்களை  பார்க்கலாம். *******") 
        else:
            print(u"******* Goodbye! Now have a nice day *******")
        if doExit:
            sys.exit( 0 )
        return

# program name / CLI args handler
def get_prog_name(lang):
    prog_name=None
    debug=False
    parser = argparse.ArgumentParser(prog=lang)
    parser.add_argument(u"files",nargs='*',default=[])
    parser.add_argument(u"-debug",action=u"store_true",
                        default=False,
                        help=u"enable debugging information on screen")
    parser.add_argument(u"-tamilencoding",default=u"UTF-8",
                        help=u"option to specify other file encodings; supported  encodings are TSCII, and UTF-8")
    parser.add_argument(u"-profile",action=u"store_true",
                        default=None,
                        help=u"profile the input file(s)")
    parser.add_argument(u"-stdin",action=u"store_true",
                        default=None,
                        help=u"read input from the standard input")
    parser.add_argument(u"-stacksize",
                        default=128,
                        help=u"default stack size for the Interpreter")
    args = parser.parse_args()

    # do it well w.r.t str names    
    if not PYTHON3:
        for itr in range(0,len(args.files)):
            args.files[itr] = args.files[itr].decode("utf-8")
    
    if args.debug:
        print(u"===>",args.tamilencoding,u" | ===> files: ".join(args.files))
    error_fcn = lambda : parser.print_help() or sys.exit(-1)
    if (len(args.files) == 0 and (not args.stdin)):
        error_fcn()
    if not(args.tamilencoding in ["utf-8","tscii","TSCII","UTF-8"]):
        print(u"ERROR: Unsupported encoding %s; use values TSCII or UTF-8, or see Help"%args.tamilencoding)
        error_fcn()
    prog_name = args.files
    debug = args.debug
    dostdin = args.stdin
    encoding = args.tamilencoding
    profile = args.profile and True or False
    stacksize = int(args.stacksize)
    if dostdin and profile:
        print(u"ERROR: Cannot run profiling and input from console")
        error_fcn()
    if args.debug:
        print(u"###### chosen encoding => ",encoding,stacksize)
    return [prog_name, debug, dostdin,encoding,stacksize,profile]

# tamil/Interpreter.py

class Interpreter:
    def __init__(self):
        # your init code
        pass

    def execute(self, source_code):
        # your logic to parse & run Tamil code
        print("Executing Tamil code:")
        print(source_code)

# ✅✅✅ Add this BELOW your class!
def main(filename):
    with open(filename, encoding="utf-8") as f:
        source_code = f.read()

    interpreter = Interpreter()  # ✅ create your interpreter
    interpreter.execute(source_code)  # ✅ run it!



#!/usr/bin/python
## -*- coding: utf-8 -*-
## 
## (C) 2008, 2013 Muthiah Annamalai
## Licensed under GPL Version 3
## 
## This module is the pretty-printer for the tamil language.
## It contains classes @WikiStyle, @Printer
## 
from __future__ import print_function
from .theme import XsyTheme
from .tamil_scanner import tamilLex, tamilToken
from .tamil import tamilInterpreter
from .transform import Visitor

class WikiStyle:
    @staticmethod
    def wrap_msg(attrib,text):
        """ @text is any character stream that has to be wrapper in a style-Theme,
            specified by @attrib """
        if ( len(attrib) >= 1 ):
            out = u'<span style="color:#'+attrib[0]
        if ( len(attrib) >= 2 ):
            out = out + u';background:#'+attrib[1]
        out = out + u'">' + attrib.process(text) + u"</span>"
        return out

class Printer(Visitor):
    def __init__(self,src_file):
        """ @styler uses a Wiki/HTML/LaTeX output formatter, with a color theme 
        specificied by @theme to render the text into the appropriate format"""
        Visitor.__init__(self)
        self.styler = WikiStyle.wrap_msg
        self.theme = XsyTheme()
        self.lexer = tamilLex(src_file)
        #print self.lexer.comments.keys()
        self.output = []
        self.line = 1 #current line 
        self.NEWLINE = "<BR />\n"
        
    def update_line(self,obj):
        #print obj.line,"= line"
        if ( obj.line > self.line ):
            self.line = obj.line
            self.append( self.NEWLINE  )
        if ( self.lexer.comments.has_key(self.line) ):
            #print "visiting comment "
            self.append( self.lexer.comments[self.line] )
            del self.lexer.comments[self.line]
        return
    
    def append(self,string):
        self.output.append( string )
    
    def default(self,*args):
        """ /dev/zero dump for all visitor methods when not handled in derived class"""
        #args[0] is AST object
        self.append(u"def :")
        self.update_line(args[0])
        self.append(str(args[0]))
        
    def visit_identifier(self, IDobj):  
        attrib = self.theme.Variables
        self.update_line(IDobj)
        self.append( self.styler(attrib,str(IDobj.id)))
        return
    
    def visit_string(self, string):
        attrib = self.theme.LiteralString
        self.update_line(string)
        self.append(self.styler(attrib,str(string)))
        return

    def visit_number(self, num):
        attrib = self.theme.LiteralNumber
        self.update_line(num)
        self.append(self.styler(attrib,str(num)))
        return

    def visit_expr_call(self,expr_call):
        var_attrib = self.theme.Variables
        self.update_line(expr_call)
        self.append(self.styler(var_attrib,str(expr_call.func_id.id)))
        op_attrib = self.theme.Operators
        self.append( self.styler(op_attrib,"(") )
        expr_call.arglist.visit( self )
        self.append( self.styler(op_attrib,")") )
        return

    def visit_expr_list(self, expr_list):
        for pos,exp_itr in enumerate(expr_list.exprs):
            self.update_line(exp_itr)
            exp_itr.visit( self )
            if (pos+1) < len(expr_list.exprs):
                self.append(",")
            
        return
    
    def visit_stmt_list(self,stmt_list):
        for stmt in stmt_list.List:
            stmt.visit(self)
            self.update_line(stmt)
            self.append(self.NEWLINE)
        return
    
    def visit_stmt( self, stmt):
        ## is this a recipe for getting stuck in a loop?
        self.update_line(stmt)
        stmt.visit(self)
        return
    
    def visit_binary_expr(self,expr):
        self.visit_expr(expr)
        return
        
    def visit_expr(self, expr):
        op_attrib = self.theme.Operators
        self.update_line(expr)
        expr.term.visit(self)
        self.append( self.styler(op_attrib, tamilToken.token_types[expr.binop.kind] ) )
        expr.next_expr.visit(self)
        return
    
    def visit_return_stmt(self, ret_stmt):
        kw_attrib = self.theme.Keywords
        self.update_line(ret_stmt)
        keyword = u"பின்கொடு"
        self.append( self.styler( kw_attrib, keyword ) )
        # return may have optional argument
        if hasattr(ret_stmt.rvalue,'visit'):
            ret_stmt.rvalue.visit(self)
        self.append(self.NEWLINE)
        return
    
    def visit_break_stmt(self, break_stmt ):
        kw_attrib = self.theme.Keywords
        self.update_line(break_stmt)
        keyword = u"நிறுத்து" #tamilToken.Keywords["break"]
        self.append( self.NEWLINE )
        self.append( self.styler( kw_attrib, keyword ) )
        self.append( self.NEWLINE )
        return
    
    def visit_continue_stmt(self, cont_stmt):
        kw_attrib = self.theme.Keywords
        self.update_line(cont_stmt)
        keyword = u"தொடர்" #tamilToken.Keywords["continue"]
        self.append( self.styler( kw_attrib, keyword ) )
        self.append( self.NEWLINE )        
        return
    
    def visit_else_stmt(self,else_stmt):
        kw_attrib = self.theme.Keywords
        keyword = u"இல்லை"
        self.append( self.styler( kw_attrib, keyword ) )
        self.update_line(else_stmt)
        else_stmt.stmt.visit( self )
        return
    
    def visit_if_elseif_stmt(self,if_elseif_stmt):
        op_attrib = self.theme.Operators
        
        # condition expression
        self.append( self.styler( op_attrib, "@( " ) )
        if_elseif_stmt.expr.visit(self)
        self.append( self.styler( op_attrib, ") " ) )
        
        # IF kw
        kw_attrib = self.theme.Keywords
        keyword_if = u"ஆனால்"
        self.append( self.styler( kw_attrib, keyword_if ) )                
        
        # True-Body
        if_elseif_stmt.body.visit( self )
        
        # False-Body - optionally present
        if hasattr(if_elseif_stmt.next_stmt,'visit'):
            if_elseif_stmt.next_stmt.visit(self)
        
        self.visit_end_kw()
        
    def visit_end_kw(self):
        # END kw
        kw_attrib = self.theme.Keywords
        keyword_end = u"முடி"
        self.append( self.styler( kw_attrib, keyword_end ) )
        
    def visit_while_stmt(self,while_stmt):
        """
        @( itr < L ) வரை
                        சமம்= சமம் + input[itr]*wts[itr]
            itr = itr + 1
                முடி"""        
        op_attrib = self.theme.Operators
        
        # condition expression
        self.append( self.styler( op_attrib, u"@( " ) )
        while_stmt.expr.visit(self)
        self.append( self.styler( op_attrib, u") " ) )
        
        # While kw
        kw_attrib = self.theme.Keywords
        keyword_while = u"வரை"
        self.append( self.styler(kw_attrib,(keyword_while)) )
        
        # Body
        while_stmt.body.visit( self )
        
        self.visit_end_kw()
        return
    
    # foreach is transformed at the AST-level
    # so its really a MACRO here
    def visit_for_stmt(self,for_stmt):
        """
        @( x = -1 , x < 0, x = x + 1 ) ஆக
                        பதிப்பி x, "கருவேபில"
                முடி
        """
        op_attrib = self.theme.Operators
        
        # condition expression
        self.append( self.styler( op_attrib, u"@( " ) )
        for_stmt.expr_init.visit(self)
        self.append( self.styler( op_attrib, u", " ) )
        for_stmt.expr_cond.visit(self)
        self.append( self.styler( op_attrib, u", " ) )
        for_stmt.expr_update.visit(self)
        self.append( self.styler( op_attrib, u") " ) )
        
        # For kw
        kw_attrib = self.theme.Keywords
        keyword_for = u"ஆக"
        self.append( self.styler(kw_attrib,keyword_for) )
        self.append( self.NEWLINE )
        # Body
        for_stmt.body.visit( self )
        self.visit_end_kw()
        return

    def visit_assign_stmt(self, assign_stmt):
        op_attrib = self.theme.Operators
        assign_stmt.lvalue.visit( self )
        self.append(self.styler(op_attrib,"="))        
        assign_stmt.rvalue.visit( self )
        return

    def visit_print_stmt(self, print_stmt):
        kw_attrib = self.theme.Keywords
        keyword = u"பதிப்பி"
        self.update_line(print_stmt)
        self.append( self.styler(kw_attrib,str(keyword)) )
        print_stmt.exprlst.visit(self)
        return

    def visit_eval_stmt(self, eval_stmt ):
        eval_stmt.expr.visit(self)
        return
    
    def visit_arg_list(self, arg_list):
        L = len(arg_list.get_list())
        for pos,arg in enumerate(arg_list.get_list()):
            if hasattr(arg,'visit'):
                arg.visit(self)
            else:
                var_attrib = self.theme.Variables
                self.append( self.styler(var_attrib,str(arg)) )            
            if ( pos < (L-1) ):
                self.append( ", " )
        return

    def visit_value_list(self,value_list):
        op_attrib = self.theme.Operators
        for value in value_list.args:
            self.update_line(value)
            value.visit(self)
            self.append(self.styler(op_attrib,","))
        self.output.pop()
        return

    def visit_function(self,fndecl_stmt):
        """
        நிரல்பாகம் fibonacci_தமிழ்( x )
        @( x <= 1 ) ஆனால்
                        ஈ = 1
                இல்லை
                        ஈ = fibonacci_தமிழ்( x - 1 ) + fibonacci_தமிழ்( x - 2 )
                முடி 
                பின்கொடு ஈ
        முடி
        """
        
        # Function kw
        kw_attrib = self.theme.Keywords
        keyword_fn = u"நிரல்பாகம்"
        self.append( self.styler(kw_attrib,keyword_fn) )
        
        # name of function
        fn_attrib = self.theme.Variables
        self.append( self.styler( fn_attrib, fndecl_stmt.name ) )
        
        # arglist expression
        op_attrib = self.theme.Operators
        self.append( self.styler( op_attrib, u"( " ) )
        fndecl_stmt.arglist.visit(self)
        self.append( self.styler( op_attrib, u") " ) )
        
        # Body expression
        fndecl_stmt.body.visit(self)
        
        return
    
    def pretty_print(self):
        self.parse_eval = tamilInterpreter(lexer=self.lexer)
        ast = self.parse_eval.parse()
        print(str(ast))
        ast.visit(self)
        
        # dump remaining comments
        comm_attrib = self.theme.Comment
        for  line,comment in  self.lexer.comments.items():
            self.append( self.styler( comm_attrib, comment ) )
            self.append( self.NEWLINE )
        
        return u"".join(self.output)
        
    # method walks the lexer-tokens and calls the appropriate elements
    # basic lexical hiliting. This is useful when parsing fails, or just
    # a plain vanilla hiliter
    def lexical_hilite(self):
        self.lexer.tokens.reverse()
        out = []
        for t in self.lexer.tokens:
            add_br = False
            attrib = self.theme.Operators
            if tamilToken.is_keyword(t.kind):
                attrib = self.theme.Keywords
                if ( tamilToken.get_name(t.kind) in ["END", "ELSE"] ):
                    out.append(u'<BR />\n')
            elif tamilToken.is_number(t.kind):
                attrib = self.theme.LiteralNumber
            elif tamilToken.is_string(t.kind):
                attrib = self.theme.LiteralString
                t.val = u'"'+t.val+u'"' #FIXME: ideally do some escaping as well
            elif tamilToken.is_id(t.kind):
                attrib = self.theme.Variables
            elif( t.val in [u"@", u"பதிப்பி" ] ):
                attrib = self.theme.Builtins
                out.append(u'<BR />\n')
            
            t.val = u" " + str(t.val)
            out.append( self.styler(attrib,t.val) )
            if ( add_br ):
                out.append(u"<BR />\n")
        
        return u"".join(out)

def pretty_print_file( filename ):
    print(u"working with ",filename)
    print(Printer(filename).pretty_print())
    return True

if __name__ == "__main__":
    from sys import argv,exit
    if len(argv) <= 1:
        print(u"usage: python tamil/prettify.py <file1> <file2> ... ")
        exit(-1)
    map( pretty_print_file, argv[1:])
    


# tamil/__main__.py

import sys
from tamil.Interpreter import main as tamil_main

def main():
    if len(sys.argv) < 2:
        print("❌ கோப்பு பெயரை கொடுக்கவும்! (Please provide a Tamil source file.)")
        print("உதாரணம் (Example): python -m tamil myfile.tamil")
        sys.exit(1)

    filename = sys.argv[1]

    try:
        tamil_main(filename)
    except Exception as e:
        print(f"⚡ கோப்பை இயக்க முடியவில்லை! (Error running file: {e})")
        sys.exit(1)

if __name__ == "__main__":
    main()
